from __future__ import annotations

import asyncio
import os
import threading
import time
import warnings

from enum import Enum
from typing import Any, Iterator

import structlog
import zmq
import zmq.asyncio

from zmq.utils.monitor import parse_monitor_message

from ...config import global_config
from ...constants import ZMQ_EVENT_MAP, ZMQ_TRANSPORTS
from ...serializers.serializers import Serializers
from ...utils import (
    format_exception_as_json,
    get_current_async_loop,
    get_sanitized_filename_from_random_string,
    get_socket_type_name,
    run_callable_somehow,
    uuid_hex,
)
from ..exceptions import BreakLoop
from .message import (
    ERROR,
    EXIT,
    HANDSHAKE,
    INVALID_MESSAGE,
    REPLY,
    SERVER_DISCONNECTED,
    TIMEOUT,
    EventMessage,
    PreserializedData,
    PreserializedEmptyByte,
    RequestMessage,
    ResponseMessage,
    SerializableData,
    SerializableNone,
    ServerExecutionContext,
    ThingExecutionContext,
    default_server_execution_context,
    default_thing_execution_context,
)


class BaseZMQ:
    """
    Base class for all ZMQ message brokers. Implements socket creation & logger config,
    which are common to all server and client implementations.
    """

    def __init__(self, id: str, **kwargs) -> None:
        """
        id: str
            unique ID of the server/client. This is used as the identity of the ZMQ socket.
        logger: logging.Logger, optional
            logger instance to use. If None, a default logger is created.
        """
        super().__init__()
        self.id = id  # type: str
        self.context = self.context if hasattr(self, "context") and self.context else None  # type: zmq.Context | zmq.asyncio.Context
        self.socket = self.socket if hasattr(self, "socket") and self.socket else None  # type: zmq.Socket | None
        self.socket_address = self.socket_address if hasattr(self, "socket_address") and self.socket_address else None  # type: str | None

    def exit(self) -> None:
        """
        Cleanup method to terminate ZMQ sockets and contexts before quitting. Called by `__del__()`
        automatically. Each subclass server/client should implement their version of exiting if necessary.
        """
        if not hasattr(self, "logger") or not self.logger:
            self.logger = structlog.get_logger().bind(component="broker", impl=self.__class__.__name__, id=self.id)

    def __del__(self) -> None:
        self.exit()

    @classmethod
    def get_socket(
        cls,
        *,
        server_id: str,
        socket_id: str,
        node_type: str,
        context: zmq.asyncio.Context | zmq.Context,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        socket_type: zmq.SocketType = zmq.ROUTER,
        **kwargs,
    ) -> tuple[zmq.Socket, str]:
        """
        Create a socket with certain specifications. Supported ZeroMQ transports are TCP, IPC & INPROC.
        For IPC sockets, a file is created under TEMP_DIR of global configuration.

        Parameters
        ----------
        server_id: str
            Used to create socket address
        socket_id: str
            Each ROUTER socket require unique identity to correctly route the messages, usually same as `server_id`
        node_type: str
            server or client? i.e. whether to bind (server) or connect (client) as per ZMQ definition
        context: zmq.Context or zmq.asyncio.Context
            ZeroMQ Context object that creates the socket
        access_point: Enum
            `TCP`, `IPC` or `INPROC`. Message crafting-passing-routing is transport-invariant as suggested by ZMQ.
            Speed relationship - `INPROC` > `IPC` > `TCP`. For client side `TCP`, specify the TCP address - `tcp://<host>:<port>`
            or a random port will be chosen.
        socket_type: zmq.SocketType, default zmq.ROUTER
            Usually a ROUTER socket is implemented for both client-server and peer-to-peer communication. But other sockets
            like PAIR, DEALER, etc. can also be used as per the use-case.
        kwargs: dict
            Additional arguments:

            - `socket_class`: `zmq.Socket` class to use, default is `zmq.Socket` for sync and `zmq.asyncio.Socket` for async.

        Returns
        -------
        socket: zmq.Socket
            created socket
        socket_address: str
            qualified address of the socket created for any transport type

        Raises
        ------
        NotImplementedError
            if transport other than `TCP`, `IPC` or `INPROC` is used
        RuntimeError
            if transport is `TCP` and a socket connection from client side is requested but a socket address is not supplied
        """
        if node_type.lower() not in ["server", "client"]:
            raise ValueError(f"Invalid node_type: {node_type}")
        bind = node_type.lower() == "server"
        if len(access_point) == 3 or len(access_point) == 6 or isinstance(access_point, Enum):
            transport = access_point
            socket_address = None
        else:
            transport = access_point.split("://")[0].upper()
            socket_address = access_point

        socket = context.socket(socket_type, socket_class=kwargs.get("socket_class", None))
        socket.setsockopt_string(zmq.IDENTITY, socket_id)

        if transport == ZMQ_TRANSPORTS.IPC or transport.lower() == "ipc":
            if socket_address is None or not socket_address.endswith(".ipc"):
                if not socket_address:
                    filename = get_sanitized_filename_from_random_string(server_id, "ipc")
                elif not socket_address.endswith(".ipc"):
                    filename = get_sanitized_filename_from_random_string(socket_address, "ipc")
                # re-compute for IPC because it looks for a file in a directory
                filename = filename.replace(
                    EventPublisher._standard_address_suffix.replace("/", "_"),
                    f".{EventPublisher._standard_address_suffix_filename_replacement}",
                )
                socket_address = "ipc://{}{}{}".format(global_config.TEMP_DIR_SOCKETS, os.sep, filename)
            if bind:
                socket.bind(socket_address)
            else:
                socket.connect(socket_address)
        elif transport == ZMQ_TRANSPORTS.TCP or transport.lower() == "tcp":
            if bind:
                failed = False
                if socket_address:
                    try:
                        socket.bind(socket_address)
                    except zmq.error.ZMQError as ex:
                        if not ex.strerror.startswith("Address in use"):
                            raise ex from None
                        failed = True
                if failed or not socket_address:
                    for i in range(
                        global_config.TCP_SOCKET_SEARCH_START_PORT,
                        global_config.TCP_SOCKET_SEARCH_END_PORT,
                    ):
                        socket_address = "tcp://0.0.0.0:{}".format(i)
                        try:
                            socket.bind(socket_address)
                            break
                        except zmq.error.ZMQError as ex:
                            if not ex.strerror.startswith("Address in use"):
                                raise ex from None
            elif socket_address:
                socket.connect(socket_address)
            else:
                raise RuntimeError(f"Socket address not supplied for TCP connection to identity - {socket_id}")
        elif transport == ZMQ_TRANSPORTS.INPROC or transport.lower() == "inproc":
            # inproc_id = id.replace('/', '_').replace('-', '_')
            if socket_address is None:
                socket_address = f"inproc://{server_id}"
            elif not socket_address.startswith("inproc://"):
                socket_address = f"inproc://{socket_address}"
            if bind:
                socket.bind(socket_address)
            else:
                socket.connect(socket_address)
        else:
            raise NotImplementedError(
                "transports other than IPC, TCP & INPROC are not implemented now for {}.".format(cls.__name__)
                + f" Given transport {transport}."
            )
        return socket, socket_address


class BaseAsyncZMQ(BaseZMQ):
    """
    Base class for all async ZMQ servers and clients.
    """

    # init of this class must always take empty arguments due to inheritance structure

    def create_socket(
        self,
        *,
        server_id: str,
        socket_id: str,
        node_type: str = "server",
        context: zmq.asyncio.Context | None = None,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        socket_type: zmq.SocketType = zmq.ROUTER,
        **kwargs,
    ) -> None:
        """
        Overloads `create_socket()` to create, bind/connect an async socket. A global context is used if
        none is supplied.
        """
        if context and not isinstance(context, zmq.asyncio.Context):
            raise TypeError(
                "async ZMQ message broker accepts only async ZMQ context. supplied type {}".format(type(context))
            )
        self.context = context or global_config.zmq_context()
        self.socket, self.socket_address = BaseZMQ.get_socket(
            server_id=server_id,
            socket_id=socket_id,
            node_type=node_type,
            context=self.context,
            access_point=access_point,
            socket_type=socket_type,
            **kwargs,
        )
        self.logger.info(
            "created socket type: {} with address: {} & identity: {} and {}".format(
                get_socket_type_name(socket_type),
                self.socket_address,
                socket_id,
                "bound" if node_type == "server" else "connected",
            )
        )


class BaseSyncZMQ(BaseZMQ):
    """Base class for all sync ZMQ servers and clients"""

    # init of this class must always take empty arguments due to inheritance structure

    def create_socket(
        self,
        *,
        server_id: str,
        socket_id: str,
        node_type: str = "server",
        context: zmq.Context | None = None,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        socket_type: zmq.SocketType = zmq.ROUTER,
        **kwargs,
    ) -> None:
        """
        Overloads `create_socket()` to create, bind/connect a synchronous socket. A global context is used if
        none is supplied.
        """
        self.context = context or global_config.zmq_context()
        self.socket, self.socket_address = BaseZMQ.get_socket(
            server_id=server_id,
            socket_id=socket_id,
            node_type=node_type,
            context=self.context,
            access_point=access_point,
            socket_type=socket_type,
            socket_class=zmq.Socket,
            **kwargs,
        )
        self.logger.info(
            "created socket type: {} with address: {} & identity: {} and {}".format(
                get_socket_type_name(socket_type),
                self.socket_address,
                socket_id,
                "bound" if node_type == "server" else "connected",
            )
        )


class BaseZMQServer(BaseZMQ):
    """Base class for all ZMQ servers irrespective of sync and async"""

    def __init__(self, id: str, logger: structlog.stdlib.BoundLogger | None = None, **kwargs) -> None:
        super().__init__(id=id, **kwargs)
        if not logger:
            logger = structlog.get_logger()
        self.logger = logger.bind(component="broker", impl=self.__class__.__name__, id=self.id)

    def handshake(self, request_message: RequestMessage) -> None:
        """
        Pass a handshake message to client. Absolutely mandatory to handshake with all clients to ensure initial messages
        do not get lost because of ZMQ's tiny but significant initial delay after creating socket.

        Parameters
        ----------
        request_message: RequestMessage
            the client message for which the handshake is being sent
        """
        run_callable_somehow(self._handshake(request_message))

    def _handshake(self, request_message: RequestMessage) -> None:
        raise NotImplementedError(
            f"handshake cannot be handled - implement _handshake in {self.__class__} to handshake."
        )

    def handle_invalid_message(self, request_message: RequestMessage, exception: Exception) -> None:
        """
        Pass an invalid message to the client when an exception occurred while parsing the message from the client
        (in `handled_default_message_types()`)

        Parameters
        ----------
        request_message: RequestMessage
            the client message for which the parsing exception occurred
        exception: Exception
            exception object raised
        """
        run_callable_somehow(self._handle_invalid_message(request_message, exception))

    def _handle_invalid_message(self, message: RequestMessage, exception: Exception) -> None:
        raise NotImplementedError(
            "invalid message cannot be handled"
            + f" - implement _handle_invalid_message in {self.__class__} to handle invalid messages."
        )

    def handle_timeout(self, request_message: RequestMessage, timeout_type: str) -> None:
        """
        Pass timeout message to the client when the operation could not be executed within specified timeouts

        Parameters
        ----------
        request_message: RequestMessage
            the client message which could not executed within the specified timeout. timeout value is
            generally specified within the execution context values.
        """
        run_callable_somehow(self._handle_timeout(request_message, timeout_type=timeout_type))

    def _handle_timeout(self, request_message: RequestMessage, timeout_type: str) -> None:
        raise NotImplementedError(
            "timeouts cannot be handled ",
            f"- implement _handle_timeout in {self.__class__} to handle timeout.",
        )

    def handle_error_message(self, request_message: RequestMessage, exception: Exception) -> None:
        """
        Pass an exception message to the client when an exception occurred while executing the operation

        Parameters
        ----------
        request_message: RequestMessage
            the client message for which the exception occurred
        exception: Exception
            exception object raised
        """
        run_callable_somehow(self._handle_error_message(request_message, exception))

    def _handle_error_message(self, request_message: RequestMessage, exception: Exception) -> None:
        raise NotImplementedError(
            "exceptions cannot be handled ",
            f"- implement _handle_error_message in {self.__class__} to handle exceptions.",
        )

    def handled_default_message_types(self, request_message: RequestMessage) -> bool:
        """
        Handle default cases for the server without further processing of the request (for example, `HANDSHAKE`).
        This method is called once/supposed to be called when the message is received or popped out of the socket.

        Parameters
        ----------
        request_message: RequestMessage
            the client message to handle
        """
        if request_message.type == HANDSHAKE:
            self.handshake(request_message)
            return True
        elif request_message.type == EXIT:
            # self.send response with message type EXIT
            raise BreakLoop(f"exit message received from {request_message.sender_id} with msg-ID {request_message.id}")
        elif request_message.length != len(request_message.byte_array):
            self.handle_invalid_message(
                request_message,
                exception=ValueError(
                    f"message length {request_message.length} does not match "
                    + f"the number of message parts {len(request_message.byte_array)}"
                ),
            )
            return True
        return False


class AsyncZMQServer(BaseZMQServer, BaseAsyncZMQ):
    """
    An async ZMQ server that can handle multiple clients. Implements both blocking (non-polled) and
    non-blocking/polling form of receiving messages and sending replies. This server can be stopped from server side
    by calling `stop_polling()`.
    """

    def __init__(
        self,
        *,
        id: str,
        context: zmq.asyncio.Context | None = None,
        socket_type: zmq.SocketType = zmq.ROUTER,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        poll_timeout: int = 25,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            `id` of the server, also used as the identity of the ZMQ socket
        context: Optional, zmq.asyncio.Context
            ZeroMQ Context object to use. If None, a global context is used.
        socket_type: zmq.SocketType, default zmq.ROUTER
            socket type of ZMQ socket, default is ROUTER (enables address based routing of messages)
        access_point: Enum | str, default ZMQ_TRANSPORTS.IPC
            Use `TCP` or `tcp://*:<port>` for network access, `IPC` for multi-process applications,
            and `INPROC` for multi-threaded applications.
        poll_timeout: int, default 25
            time in milliseconds to poll the sockets specified under `procotols`. Useful for calling `stop_polling()`
            where the max delay to stop polling will be `poll_timeout`.
        kwargs: dict
            Additional arguments for `BaseZMQ` and `BaseAsyncZMQ` classes.

            - `logger`: logger instance to use. If None, a default logger is created.
        """

        super().__init__(id=id, **kwargs)
        self.create_socket(
            server_id=id,
            socket_id=id,
            node_type="server",
            context=context,
            access_point=access_point,
            socket_type=socket_type,
            **kwargs,
        )  # for server the server ID and socket ID is the same, only for clients they differ
        self.poller = zmq.asyncio.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self.poll_timeout = poll_timeout

    @property
    def poll_timeout(self) -> int:
        """socket polling timeout in milliseconds greater than 0"""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, value) -> None:
        if not isinstance(value, int) or value < 0:
            raise ValueError(
                f"polling period must be an integer greater than 0, not {value}. "
                + "Value is considered in milliseconds."
            )
        self._poll_timeout = value

    async def async_recv_request(self) -> RequestMessage:
        """
        Receive one message in a blocking form. There is no polling, therefore this method blocks until a message is
        received.

        Returns
        -------
        message: RequestMessage
            received message
        """
        while True:
            raw_message = await self.socket.recv_multipart()
            request_message = RequestMessage(raw_message)
            if not self.handled_default_message_types(request_message) and raw_message:
                self.logger.debug(
                    "received message from client",
                    client_id=request_message.sender_id,
                    msg_id=request_message.id,
                    message_type=request_message.type,
                )
                return request_message

    async def async_recv_requests(self) -> list[RequestMessage]:
        """
        Receive all currently available messages in blocking form. There is no polling, therefore this method
        blocks until at least one message is received.

        Returns
        -------
        messages: list[RequestMessage]
            list of received messages
        """
        messages = [await self.async_recv_request()]
        while True:
            try:
                raw_message = await self.socket.recv_multipart(zmq.NOBLOCK)
                request_message = RequestMessage(raw_message)
                if not self.handled_default_message_types(request_message) and raw_message:
                    self.logger.debug(
                        "received message from client",
                        client_id=request_message.sender_id,
                        msg_id=request_message.id,
                        message_type=request_message.type,
                    )
                    messages.append(request_message)
            except zmq.Again:
                break
        return messages

    async def async_send_response(
        self,
        request_message: RequestMessage,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> None:
        """
        Send response message for a request message.

        Parameters
        ----------
        request_message: RequestMessage
            original message based on which the response message can be automatically crafted and routed
        payload: SerializableData
            serializable data to be sent as response
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        """
        response_message = ResponseMessage.craft_reply_from_request(
            request_message=request_message,
            payload=payload,
            preserialized_payload=preserialized_payload,
        )
        await self.socket.send_multipart(response_message.byte_array)
        self.logger.debug(
            "sent response to client",
            receiver_id=response_message.receiver_id,
            msg_id=response_message.id,
            message_type=response_message.type,
        )

    async def async_send_response_with_message_type(
        self,
        request_message: RequestMessage,
        message_type: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> None:
        """
        Send a specific response message apart from `REPLY` for a request message.

        Parameters
        ----------
        request_message: RequestMessage
            original message based on which the response message can be automatically crafted and routed
        payload: SerializableData
            serializable data to be sent as response
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        """
        response_message = ResponseMessage.craft_from_arguments(
            receiver_id=request_message.sender_id,
            sender_id=self.id,
            message_type=message_type or REPLY,
            message_id=request_message.id,
            payload=payload,
            preserialized_payload=preserialized_payload,
        )
        await self.socket.send_multipart(response_message.byte_array)
        self.logger.debug(
            "sent response to client",
            receiver_id=response_message.receiver_id,
            msg_id=response_message.id,
            message_type=response_message.type,
        )

    async def poll_requests(self) -> list[RequestMessage]:
        """
        poll for messages with specified timeout (`poll_timeout`) and return if any messages are available.
        This method can be stopped from another method in a different thread or asyncio task (not in the same thread though).

        Returns
        -------
        messages: list[RequestMessage]
            list of received messages
        """
        self.stop_poll = False
        messages = []
        while not self.stop_poll:
            sockets = await self.poller.poll(self._poll_timeout)  # type hints dont work in this line
            for socket, _ in sockets:
                while True:
                    try:
                        raw_message = await socket.recv_multipart(zmq.NOBLOCK)
                    except zmq.Again:
                        break
                    else:
                        request_message = RequestMessage(raw_message)
                        if not self.handled_default_message_types(request_message) and raw_message:
                            self.logger.debug(
                                "received message from client",
                                sender_id=request_message.sender_id,
                                receiver_id=request_message.receiver_id,
                                msg_id=request_message.id,
                                message_type=request_message.type,
                            )
                            messages.append(request_message)
            if len(messages) > 0:
                break
        return messages

    def stop_polling(self) -> None:
        """stop polling and unblock `poll_messages()` method"""
        self.stop_poll = True

    async def _handshake(self, request_message: RequestMessage) -> None:
        """
        Inner method that handles handshake. Scheduled by `handshake()` method, signature same as `handshake()`.
        """
        # Note that for ROUTER sockets, once the message goes through the sending socket, the address of the receiver
        # is replaced by the address of the sender once received
        handshake_message = ResponseMessage.craft_from_arguments(
            receiver_id=request_message.sender_id,
            sender_id=self.id,
            message_type=HANDSHAKE,
            message_id=request_message.id,
        )
        await self.socket.send_multipart(handshake_message.byte_array)
        self.logger.info(
            "sent handshake to client",
            receiver_id=handshake_message.receiver_id,
            msg_id=handshake_message.id,
        )

    async def _handle_timeout(self, request_message: RequestMessage, timeout_type: str) -> None:
        """
        Inner method that handles timeout. Scheduled by `handle_timeout()`, signature same as `handle_timeout`.
        """
        timeout_message = ResponseMessage.craft_from_arguments(
            receiver_id=request_message.sender_id,
            sender_id=self.id,
            message_type=TIMEOUT,
            message_id=request_message.id,
            payload=SerializableData(timeout_type, content_type="application/json"),
        )
        await self.socket.send_multipart(timeout_message.byte_array)
        self.logger.warning(
            f"sent {timeout_type} timeout to client",
            receiver_id=timeout_message.receiver_id,
            msg_id=timeout_message.id,
        )

    async def _handle_invalid_message(self, request_message: RequestMessage, exception: Exception) -> None:
        """
        Inner method that handles invalid messages. Scheduled by `handle_invalid_message()`,
        signature same as `handle_invalid_message()`.
        """
        invalid_message = ResponseMessage.craft_from_arguments(
            receiver_id=request_message.sender_id,
            sender_id=self.id,
            message_type=INVALID_MESSAGE,
            message_id=request_message.id,
            payload=SerializableData(
                dict(exception=format_exception_as_json(exception)),
                content_type="application/json",
            ),
        )
        await self.socket.send_multipart(invalid_message.byte_array)
        self.logger.warning(
            f"informed client about invalid message due to exception - {str(exception)}",
            receiver_id=invalid_message.receiver_id,
            msg_id=invalid_message.id,
        )

    async def _handle_error_message(self, request_message: RequestMessage, exception: Exception) -> None:
        error_message = ResponseMessage.craft_with_message_type(
            request_message=request_message,
            message_type=ERROR,
            payload=SerializableData(
                dict(exception=format_exception_as_json(exception)),
                content_type="application/json",
            ),
        )
        await self.socket.send_multipart(error_message.byte_array)
        self.logger.warning(
            f"sent error message to client for exception - {str(exception)}",
            receiver_id=error_message.receiver_id,
            msg_id=error_message.id,
        )

    def exit(self) -> None:
        """unregister socket from poller and terminate socket. context is not terminated as it may be shared."""
        try:
            BaseZMQ.exit(self)
            self.poller.unregister(self.socket)
            self.socket.close(0)
            self.logger.info("terminated socket of server")
        except Exception as ex:
            self.logger.warning(f"error while closing socket - {str(ex)}")


class ZMQServerPool(BaseZMQServer):
    """Implements pool of async ZMQ servers (& their sockets)"""

    def __init__(self, *, ids: list[str] | None = None, **kwargs) -> None:
        """
        ids: List[str], optional
            list of server IDs to create the server pool. If None, an empty pool is created and servers can be
            registered later using `register_server()`.
        kwargs: dict
            Additional arguments for `BaseZMQ` and `BaseAsyncZMQ` classes.

            - `logger`: logger instance to use. If None, a default logger is created.
        """
        self.context = global_config.zmq_context()
        self.poller = zmq.asyncio.Poller()
        self.pool = dict()  # type: dict[str, AsyncZMQServer]
        if ids:
            for id in ids:
                self.pool[id] = AsyncZMQServer(id=id, context=self.context, **kwargs)
            for server in self.pool.values():
                self.poller.register(server.socket, zmq.POLLIN)
        super().__init__(id=f"pool-{uuid_hex()}", **kwargs)

    def create_socket(
        self,
        *,
        id: str,
        bind: bool,
        context: zmq.asyncio.Context | zmq.Context,
        access_point: str,
        socket_type: zmq.SocketType = zmq.ROUTER,
        **kwargs,
    ) -> None:
        raise NotImplementedError("create socket not supported by ZMQServerPool")
        # we override this method to prevent socket creation. id set to pool is simply a filler
        # return super().create_socket()

    def register_server(self, server: AsyncZMQServer) -> None:
        if not isinstance(server, (AsyncZMQServer)):
            raise TypeError(f"registration possible only for subclass of AsyncZMQServer. Given type {type(server)}")
        self.pool[server.id] = server
        self.poller.register(server.socket, zmq.POLLIN)

    def deregister_server(self, server: AsyncZMQServer) -> None:
        self.poller.unregister(server.socket)
        self.pool.pop(server.id)

    @property
    def poll_timeout(self) -> int:
        """socket polling timeout in milliseconds greater than 0"""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, value) -> None:
        if not isinstance(value, int) or value < 0:
            raise ValueError(
                "polling period must be an integer greater than 0, not {}.".format(value)
                + " Value is considered in milliseconds."
            )
        self._poll_timeout = value

    async def async_recv_request(self, id: str) -> RequestMessage:
        """
        receive message for server specified by id

        Parameters
        ----------
        id: str
            id of the ZMQ server

        Returns
        -------
        RequestMessage
            received message
        """
        return await self.pool[id].async_recv_request()

    async def async_recv_requests(self, id: str) -> list[RequestMessage]:
        """
        receive all available messages for server specified by id

        Parameters
        ----------
        id: str
            id of the ZMQ server

        Returns
        -------
        list[RequestMessage]
            list of received messages
        """
        return await self.pool[id].async_recv_requests()

    async def async_send_response(
        self,
        *,
        id: str,
        request_message: RequestMessage,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
    ) -> None:
        """
        send response for a request message for server specified by id

        Parameters
        ----------
        id: str
            id of the ZMQ server
        request_message: RequestMessage
            request message for which response is being given
        payload: SerializableData
            serializable data to be sent as response
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        """
        await self.pool[id].async_send_response(
            request_message=request_message,
            payload=payload,
            preserialized_payload=preserialized_payload,
        )

    async def poll(self) -> list[RequestMessage]:
        """
        Pool for messages in the entire server pool. Use the message to identify the server by using `receiver_id` of
        the message header.

        Returns
        -------
        list[RequestMessage]
            list of received messages across all servers in the pool
        """
        self.stop_poll = False
        messages = []
        while not self.stop_poll:
            sockets = await self.poller.poll(self._poll_timeout)
            for socket, _ in sockets:
                while True:
                    try:
                        raw_message = await socket.recv_multipart(zmq.NOBLOCK)
                    except zmq.Again:
                        break
                    else:
                        if raw_message:
                            request_message = RequestMessage(raw_message)
                            self.logger.debug(
                                "received message from client",
                                sender_id=request_message.sender_id,
                                receiver_id=request_message.receiver_id,
                                msg_id=request_message.id,
                            )
                            messages.append(request_message)
        return messages

    def stop_polling(self) -> None:
        """stop polling method `poll()`"""
        self.stop_poll = True

    def __getitem__(self, key) -> AsyncZMQServer:
        return self.pool[key]

    def __iter__(self) -> Iterator[str]:
        return self.pool.__iter__()

    def __contains__(self, name: str) -> bool:
        return name in self.pool.keys()

    def exit(self) -> None:
        for server in self.pool.values():
            try:
                self.poller.unregister(server.socket)
            except Exception as ex:
                self.logger.warning(f"could not unregister poller - {str(ex)}")
            server.exit()


class BaseZMQClient(BaseZMQ):
    """Base class for all ZMQ clients irrespective of sync and async"""

    def __init__(
        self,
        *,
        id: str,
        server_id: str,
        logger: structlog.BoundLogger | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            Unique id of the client to receive messages from the server. Each client connecting to same server must
            still have unique ID.
        server_id: str
            The server id to connect to
        logger: Optional[logging.Logger]
            logger instance to use. If None, a default logger is created.
        kwargs: dict
            Additional arguments:

            - `poll_timeout`: `int`, time in milliseconds to poll the socket for messages, default is 1000 ms.
        """
        super().__init__(id=id, **kwargs)
        if not logger:
            logger = structlog.get_logger()
        self.logger = logger.bind(component="broker", impl=self.__class__.__name__, id=id, server_id=server_id)
        self.server_id = server_id
        self.socket: zmq.Socket | zmq.asyncio.Socket
        self.poller: zmq.Poller | zmq.asyncio.Poller
        self._monitor_socket = None  # type: zmq.Socket | zmq.asyncio.Socket | None
        self._response_cache = dict()
        self._poll_timeout = kwargs.get("poll_timeout", 1000)  # default to 1000 ms
        self._stop = False  # in general, stop any loop with this variable

    @property
    def poll_timeout(self) -> int:
        """socket polling timeout in milliseconds greater than 0"""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, value: int) -> None:
        if not isinstance(value, int) or value <= 0:
            raise ValueError(
                f"polling period must be an integer greater than 0, not {value}. Value is considered in milliseconds."
            )
        self._poll_timeout = value

    def exit(self) -> None:
        try:
            BaseZMQ.exit(self)
            self.poller.unregister(self.socket)
            # TODO - there is some issue here while quitting
            if self._monitor_socket is not None:
                self.poller.unregister(self._monitor_socket)
        except Exception as ex:  # noqa
            # TODO log message and undo noqa
            # raises a weird key error for some reason
            # unable to deregister from poller - <zmq.asyncio.Socket(zmq.PAIR) at 0x1c9e5028830> - KeyError
            # unable to deregister from poller - <zmq.asyncio.Socket(zmq.PAIR) at 0x1c9e502a350> - KeyError
            # unable to deregister from poller - <zmq.asyncio.Socket(zmq.PAIR) at 0x1c9e5080750> - KeyError
            # unable to deregister from poller - <zmq.asyncio.Socket(zmq.PAIR) at 0x1c9e5082430> - KeyError
            self.logger.warning(f"unable to deregister socket from poller - {str(ex)} - {type(ex).__name__}")
        try:
            if self._monitor_socket is not None:
                self._monitor_socket.close(0)
            self.socket.close(0)
            self.logger.info("terminated socket of client")
        except Exception as ex:
            self.logger.warning(
                "could not properly terminate socket or attempted to terminate an already terminated,"
                + f" exception message: {str(ex)}"
            )

    def handled_default_message_types(self, response_message: ResponseMessage) -> bool:
        """
        Handle default cases for the client. This method is called once/supposed to be called when the message
        is received or popped out of the socket.

        Parameters
        ----------
        response_message: List[ResponseMessage]
            the server message to handle
        """
        if len(response_message.byte_array) == 2:  # socket monitor message, not our message
            try:
                if ZMQ_EVENT_MAP[parse_monitor_message(response_message.byte_array)["event"]] == SERVER_DISCONNECTED:
                    raise ConnectionAbortedError("server disconnected")
                return True  # True should simply continue polling
            except RuntimeError:
                self.logger.warning(
                    "message received from monitor socket cannot be deserialized, "
                    + f"assuming its irrelevant and skipping, {response_message.byte_array}"
                )
                return True
        elif len(response_message.byte_array) != ResponseMessage.length:  # either an error or not our message
            self.logger.warning(
                f"received unknown message from server, message length: {len(response_message.byte_array)}, "
                + f"message: {response_message.byte_array}"
            )
            return True
        if response_message.type == HANDSHAKE:
            return True
        return False

    def stop(self) -> None:
        """Stop polling for messages from server. Automatically reset when polling starts again."""
        self._stop = True


class SyncZMQClient(BaseZMQClient, BaseSyncZMQ):
    """Sync ZMQ client (as-in not asyncio compatible) that connect with a ZMQ server. Implements REQ-REP pattern."""

    def __init__(
        self,
        id: str,
        server_id: str,
        context: zmq.Context | None = None,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        handshake: bool = True,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            Unique id of the client to receive messages from the server. Each client connecting to same server must
            still have unique ID.
        server_id: str
            The server id to connect to
        context: Optional, zmq.Context
            ZeroMQ Context object to use. If None, a global context is used.
        access_point: Enum | str, default ZMQ_TRANSPORTS.IPC
            Use `TCP` or `tcp://<host>:<port>` for network access, `IPC` for multi-process applications,
            and `INPROC` for multi-threaded applications.
        handshake: bool
            When true, handshake with the server first before receiving any other message and block until that handshake was
            accomplished, highly recommended.
        kwargs: dict[str, Any]
            Additional arguments:

            - `poll_timeout`: `int`. The timeout for polling the socket (in milliseconds)
            - `handshake_timeout`: `int`. The timeout for the handshake process (in milliseconds) to complete.
                If handshake does not complete within this time, an exception is raised. Only relevant if `handshake` is True.
            - `logger`: `logging.Logger`. logger instance to use. If None, a default logger is created.
        """
        super().__init__(id=id, server_id=server_id, **kwargs)
        self.create_socket(
            server_id=server_id,
            socket_id=id,
            node_type="client",
            context=context,
            access_point=access_point,
            **kwargs,
        )
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self._poller_lock = threading.Lock()
        if handshake:
            self.handshake(kwargs.pop("handshake_timeout", 60000))

    def send_request(
        self,
        thing_id: str,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> bytes:
        """
        send request message to server.

        Parameters
        ----------
        thing_id: str
            `id` of the `Thing` on which an operation is to be performed
        objekt: str
            name of property, action or event (usually only property or action)
        operation: str
            operation to be performed, like `readproperty`, `writeproperty`, `invokeaction` etc.
        payload: SerializableData
            serializable data to be sent as payload
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        server_execution_context: dict[str, Any]
            Specify server level execution context like `invokationTimeout`, `executionTimeout`, `oneway` operation etc.
        thing_execution_context: dict[str, Any]
            Specify thing level execution context like `fetchExecutionLogs` etc.

        Returns
        -------
        bytes
            a message id in bytes
        """
        request_message = RequestMessage.craft_from_arguments(
            receiver_id=self.server_id,
            sender_id=self.id,
            thing_id=thing_id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )
        self.socket.send_multipart(request_message.byte_array)
        self.logger.debug(
            "sent message to server",
            msg_id=request_message.id,
            message_type=request_message.type,
            thing_id=thing_id,
            operation=operation,
            objekt=objekt,
        )
        return request_message.id

    def recv_response(self, message_id: bytes) -> ResponseMessage:
        """
        Receives response from server. Messages are identified by message id, and out of order messages are sent to
        a cache which may be popped later. This method blocks until the expected message is received or `stop_polling()`
        is called from another thread.

        Parameters
        ----------
        message_id: bytes
            the message id of the expected response message
        """
        self._stop = False
        while not self._stop:
            if message_id in self._response_cache:
                return self._response_cache.pop(message_id)
            try:
                if not self._poller_lock.acquire(timeout=self.poll_timeout / 1000 if self.poll_timeout else -1):
                    continue
                sockets = self.poller.poll(self.poll_timeout)
                response_message = None  # type: ResponseMessage
                for socket, _ in sockets:
                    try:
                        raw_message = socket.recv_multipart(zmq.NOBLOCK)
                        response_message = ResponseMessage(raw_message)
                    except zmq.Again:
                        pass
                    if response_message:
                        if self.handled_default_message_types(response_message):
                            continue
                        if message_id != response_message.id:
                            self._response_cache[response_message.id] = response_message
                            self.logger.debug(
                                "cached response as it does not corresponding to expected ID",
                                msg_id=response_message.id,
                                expected_msg_id=message_id,
                            )
                        else:
                            self.logger.debug(
                                "received response",
                                msg_id=response_message.id,
                                message_type=response_message.type,
                            )
                            return response_message
            finally:
                try:
                    self._poller_lock.release()
                except Exception as ex:  # noqa
                    # TODO log exception message and undo noqa
                    # 1. no need to release an unacquired lock, which can happen if another thread polling
                    # put the expected message in response message cache
                    # 2. also release the lock in every iteration because a message may be added in response cache
                    # and may not return the method, which means the loop will run again and the lock needs to reacquired
                    self.logger.warning(f"could not release poller lock for recv_response - {str(ex)}")

    def execute(
        self,
        thing_id: bytes,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> ResponseMessage:
        """
        send an operation and receive the response for it.

        Parameters
        ----------
        thing_id: str
            `id` of the `Thing` on which an operation is to be performed
        objekt: str
            name of property, action or event (usually only property or action)
        operation: str
            operation to be performed, like `readproperty`, `writeproperty`, `invokeaction` etc.
        payload: SerializableData
            serializable data to be sent as payload
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        server_execution_context: dict[str, Any]
            Specify server level execution context like `invokationTimeout`, `executionTimeout`, `oneway` operation etc.
        thing_execution_context: dict[str, Any]
            Specify thing level execution context like `fetchExecutionLogs` etc.

        Returns
        -------
        ResponseMessage
            response message from server after completing the operation
        """
        message_id = self.send_request(
            thing_id=thing_id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )
        return self.recv_response(message_id=message_id)

    def handshake(self, timeout: float | int = 60000) -> None:
        """
        handshake with server before sending first message

        Parameters
        ----------
        timeout: float | int
            timeout in milliseconds to wait for handshake to complete. If handshake does not complete within this time,
            a `ConnectionError` is raised. If None, wait indefinitely until handshake completes.
        """
        self._stop = False
        start_time = time.time_ns()
        while not self._stop:
            if timeout is not None and (time.time_ns() - start_time) / 1e6 > timeout:
                raise ConnectionError(f"Unable to contact server '{self.server_id}' from client '{self.id}'")
            handshake_message = RequestMessage.craft_with_message_type(self.id, self.server_id, HANDSHAKE)
            self.socket.send_multipart(handshake_message.byte_array)
            self.logger.info("sent Handshake to server")
            if self.poller.poll(500):
                try:
                    raw_message = self.socket.recv_multipart(zmq.NOBLOCK)
                    response_message = ResponseMessage(raw_message)
                except zmq.Again:
                    pass
                else:
                    if response_message.type == HANDSHAKE:
                        self.logger.info("client handshook with server")
                        break
                    elif self.handled_default_message_types(response_message):
                        continue
                    else:
                        self._response_cache[response_message.id] = response_message
            else:
                self.logger.info("got no response for handshake")
        self._monitor_socket = self.socket.get_monitor_socket()
        self.poller.register(self._monitor_socket, zmq.POLLIN)
        # sufficient to know when server dies only while receiving messages, not continuous polling


class AsyncZMQClient(BaseZMQClient, BaseAsyncZMQ):
    """Async ZMQ client (asyncio compatible) that connect with a ZMQ server. Implements REQ-REP pattern."""

    def __init__(
        self,
        id: str,
        server_id: str,
        context: zmq.asyncio.Context | None = None,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        handshake: bool = True,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            Unique id of the client to receive messages from the server. Each client connecting to same server must
            still have unique ID.
        server_id: str
            The server id to connect to
        context: Optional, zmq.Context
            ZeroMQ Context object to use. If None, a global context is used.
        access_point: Enum | str, default ZMQ_TRANSPORTS.IPC
            Use `TCP` or `tcp://<host>:<port>` for network access, `IPC` for multi-process applications,
            and `INPROC` for multi-threaded applications.
        handshake: bool
            When true, handshake with the server first before receiving any other message and block until that handshake was
            accomplished, highly recommended.
        kwargs: dict[str, Any]
            Additional arguments:

            - `poll_timeout`: `int`. The timeout for polling the socket (in milliseconds)
            - `handshake_timeout`: `int`. The timeout for the handshake process (in milliseconds) to complete.
                If handshake does not complete within this time, an exception is raised. Only relevant if `handshake` is True.
            - `logger`: `logging.Logger`. logger instance to use. If None, a default logger is created.
        """
        super().__init__(id=id, server_id=server_id, **kwargs)
        self.create_socket(
            server_id=server_id,
            socket_id=id,
            node_type="client",
            context=context,
            access_point=access_point,
            **kwargs,
        )
        self._monitor_socket = self.socket.get_monitor_socket()
        self.poller = zmq.asyncio.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self.poller.register(self._monitor_socket, zmq.POLLIN)
        self._poller_lock = asyncio.Lock()
        self._handshake_event = asyncio.Event()
        self._handshake_event.clear()
        if handshake:
            self.handshake(kwargs.pop("handshake_timeout", 60000))

    def handshake(self, timeout: int | None = 60000) -> None:
        """
        schedules a handshake coroutine in the running event loop
        or completes handshake synchronously if no event loop is running.
        Use `handshake_complete()` async method to check if handshake is complete.

        Parameters
        ----------
        timeout: float | int
            timeout in milliseconds to wait for handshake to complete. If handshake does not complete within this time,
            a `ConnectionError` is raised. If None, wait indefinitely until handshake completes.
        """
        run_callable_somehow(self._handshake(timeout))

    async def _handshake(self, timeout: float | int | None = 60000) -> None:
        """handshake with server before sending first message"""
        self._stop = False
        if self._monitor_socket is not None and self._monitor_socket in self.poller:
            self.poller.unregister(self._monitor_socket)
        self._handshake_event.clear()
        start_time = time.time_ns()
        while not self._stop:
            if timeout is not None and (time.time_ns() - start_time) / 1e6 > timeout:
                raise ConnectionError(f"Unable to contact server '{self.server_id}' from client '{self.id}'")
            handshake_message = RequestMessage.craft_with_message_type(self.id, self.server_id, HANDSHAKE)
            await self.socket.send_multipart(handshake_message.byte_array)
            self.logger.info("sent Handshake to server")
            if await self.poller.poll(500):
                try:
                    raw_message = await self.socket.recv_multipart(zmq.NOBLOCK)
                    response_message = ResponseMessage(raw_message)
                except zmq.Again:
                    pass
                else:
                    if response_message.type == HANDSHAKE:  # type: ignore
                        self.logger.info("client handshook with server")
                        break
                    elif self.handled_default_message_types(response_message):
                        continue
                    else:
                        self._response_cache[response_message.id] = response_message
            else:
                self.logger.info("got no response for handshake")
        self.poller.register(self._monitor_socket, zmq.POLLIN)
        self._handshake_event.set()

    async def handshake_complete(self, timeout: float | int = 60000) -> None:
        """
        wait for handshake to complete

        Parameters
        ----------
        timeout: float | int
            timeout in milliseconds to wait for handshake to complete. If handshake does not complete within this time,
            a `TimeoutError` is raised.
        """
        await asyncio.wait_for(self._handshake_event.wait(), int(timeout / 1000) if timeout else None)
        if not self._handshake_event.is_set():
            raise TimeoutError(f"Handshake with server timed out after {timeout} ms")

    async def async_send_request(
        self,
        thing_id: str,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: dict[str, Any] = default_thing_execution_context,
    ) -> str:
        """
        send request message to server.

        Parameters
        ----------
        thing_id: str
            `id` of the `Thing` on which an operation is to be performed
        objekt: str
            name of property, action or event (usually only property or action)
        operation: str
            operation to be performed, like `readproperty`, `writeproperty`, `invokeaction` etc.
        payload: SerializableData
            serializable data to be sent as payload
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        server_execution_context: dict[str, Any]
            Specify server level execution context like `invokationTimeout`, `executionTimeout`, `oneway` operation etc.
        thing_execution_context: dict[str, Any]
            Specify thing level execution context like `fetchExecutionLogs` etc.

        Returns
        -------
        bytes
            a message id in bytes
        """
        request_message = RequestMessage.craft_from_arguments(
            receiver_id=self.server_id,
            sender_id=self.id,
            thing_id=thing_id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )
        await self.socket.send_multipart(request_message.byte_array)
        self.logger.debug(
            "sent message to server",
            msg_id=request_message.id,
            message_type=request_message.type,
            thing_id=thing_id,
            operation=operation,
            objekt=objekt,
        )
        return request_message.id

    async def async_recv_response(self, message_id: str) -> list[ResponseMessage]:
        """
        Receives response from server. Messages are identified by message id, and out of order messages are sent to
        a cache which may be popped later. This method blocks until the expected message is received or `stop_polling()`
        is called from another thread.

        Parameters
        ----------
        message_id: bytes
            the message id of the expected response message
        """
        self._stop = False
        while not self._stop:
            if message_id in self._response_cache:
                return self._response_cache.pop(message_id)
            try:
                try:
                    await asyncio.wait_for(
                        self._poller_lock.acquire(),
                        timeout=self.poll_timeout / 1000 if self.poll_timeout else None,
                    )
                except TimeoutError:
                    continue
                sockets = await self.poller.poll(self._poll_timeout)
                response_message = None
                for socket, _ in sockets:
                    try:
                        raw_message = await socket.recv_multipart(zmq.NOBLOCK)
                        response_message = ResponseMessage(raw_message)
                    except zmq.Again:
                        continue
                    if response_message:
                        if self.handled_default_message_types(response_message):
                            continue
                        if message_id != response_message.id:
                            self._response_cache[response_message.id] = response_message
                            self.logger.debug(
                                "cached response as it does not corresponding to expected ID",
                                msg_id=response_message.id,
                                expected_msg_id=message_id,
                            )
                        else:
                            self.logger.debug(
                                "received response",
                                msg_id=response_message.id,
                                message_type=response_message.type,
                            )
                            return response_message
            finally:
                try:
                    self._poller_lock.release()
                except Exception as ex:
                    self.logger.warning(f"could not release poller lock for async_recv_response - {str(ex)}")

    async def async_execute(
        self,
        thing_id: str,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> ResponseMessage:
        """
        send an operation and receive the response for it.

        Parameters
        ----------
        thing_id: str
            `id` of the `Thing` on which an operation is to be performed
        objekt: str
            name of property, action or event (usually only property or action)
        operation: str
            operation to be performed, like `readproperty`, `writeproperty`, `invokeaction` etc.
        payload: SerializableData
            serializable data to be sent as payload
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        server_execution_context: dict[str, Any]
            Specify server level execution context like `invokationTimeout`, `executionTimeout`, `oneway` operation etc.
        thing_execution_context: dict[str, Any]
            Specify thing level execution context like `fetchExecutionLogs` etc.

        Returns
        -------
        ResponseMessage
            response message from server after completing the operation
        """
        message_id = await self.async_send_request(
            thing_id=thing_id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )
        return await self.async_recv_response(message_id)


class MessageMappedZMQClientPool(BaseZMQClient):
    """
    Pool of async ZMQ clients, to be primarily used within protocol bindings where multiple things may be served.
    Use message ID to track responses.
    """

    def __init__(
        self,
        id: str,
        client_ids: list[str],
        server_ids: list[str],
        handshake: bool = True,
        context: zmq.asyncio.Context = None,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        poll_timeout: int = 25,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            ID of the pool, must be unique.
        client_ids: List[str]
            list of id's of clients in the pool, must be unique. Clients are created with these ID's.
            For pre-existing clients, use `register` method instead and leave this list empty.
        server_ids: List[str]
            list of id's of servers to connect to, must be same length as client_ids. For pre-existing clients that
            are connected to servers, use `register` method instead and leave this list empty.
        handshake: bool
            when true, handshake with the server first before allowing first message and block until that handshake was
            accomplished.
        context: zmq.asyncio.Context
            ZMQ context to use, if None, a global context is used.
        access_point: Enum | str, default ZMQ_TRANSPORTS.IPC
            Default access point for all clients in the pool, usually not helpful except `INPROC` transport.
            Use `TCP` or `tcp://<host>:<port>` for network access, `IPC` for multi-process applications,
            and `INPROC` for multi-threaded applications.
        poll_timeout: int
            socket polling timeout in milliseconds greater than 0.
        **kwargs:
            Additional arguments:

            - `logger`: `logging.Logger`, logger instance to use. If None, a default logger is created.
        """
        super().__init__(id=id, server_id=None, **kwargs)
        if len(client_ids) != len(server_ids):
            raise ValueError("client_ids and server_ids must have same length")
        # this class does not call create_socket method
        self.context = context or global_config.zmq_context()
        self.pool = dict()  # type: dict[str, AsyncZMQClient]
        self.poller = zmq.asyncio.Poller()
        for client_id, server_id in zip(client_ids, server_ids):
            client = AsyncZMQClient(
                id=client_id,
                server_id=server_id,
                handshake=handshake,
                context=self.context,
                access_point=access_point,
                logger=self.logger,
            )
            self.register(client)
        # Both the client pool as well as the individual client get their serializers and client_types
        # This is required to implement pool level sending and receiving messages like polling of pool of sockets
        self.event_pool = AsyncioEventPool(len(server_ids))
        self.events_map = dict()  # type: dict[bytes, asyncio.Event]
        self.message_map = dict()
        self.cancelled_messages = []
        self.poll_timeout = poll_timeout
        self.stop_poll = False
        self._thing_to_client_map = dict()  # type: dict[str, AsyncZMQClient]
        self._client_to_thing_map = dict()  # type: dict[str, str]

    def create_new(self, id: str, server_id: str, access_point: str = ZMQ_TRANSPORTS.IPC) -> None:
        """
        Create new server with specified transport & add it to the pool.
        Other arguments are taken from pool default specifications.

        Parameters
        ----------
        id: str
            id of the new client to be created
        server_id: str
            id of the server to connect to
        access_point: str
            transport method used by the server - `IPC`, `INPROC` or `tcp://<host>:<port>`
        """
        if server_id not in self.pool.keys():
            client = AsyncZMQClient(
                id=id,
                server_id=server_id,
                handshake=True,
                context=self.context,
                access_point=access_point,
                logger=self.logger,
            )
            client._monitor_socket = client.socket.get_monitor_socket()
            self.poller.register(client._monitor_socket, zmq.POLLIN)
            self.pool[id] = client
        else:
            raise ValueError(f"client for instance name '{server_id}' already present in pool")

    def register(self, client: AsyncZMQClient, thing_id: str | None = None) -> None:
        """
        Register a pre-existing client with the pool.

        Parameters
        ----------
        client: AsyncZMQClient
            client to be registered
        thing_id: Optional, str
            thing_id to which this client is mapped, especially when the client is connected to a server that serves
            only one `Thing`.
        """
        if not isinstance(client, AsyncZMQClient):
            raise TypeError(
                "registration possible for clients only subclass of AsyncZMQClient." + f" Given type {type(client)}"
            )
        if client.id not in self.pool:
            self.pool[client.id] = client
            self.poller.register(client.socket, zmq.POLLIN)
            self.poller.register(client._monitor_socket, zmq.POLLIN)
        elif self.pool[client.id] != client:
            warnings.warn(
                f"client with id '{client.id}' already present in pool. Replacing with {client}",
                category=UserWarning,
            )
        if thing_id:
            self._thing_to_client_map[thing_id] = client.id
            self._client_to_thing_map[client.id] = thing_id

    def get_client_id_from_thing_id(self, thing_id: str) -> str:
        """
        Retrieve client mapped to a `thing_id`. The value must have been previously set using `register` method.

        Parameters
        ----------
        thing_id: str
            the `thing_id` for which the client is to be retrieved
        """
        if thing_id not in self._thing_to_client_map:
            raise ValueError(f"client for thing_id '{thing_id}' not present in pool")
        return self._thing_to_client_map.get(thing_id, None)

    def get_thing_id_from_client_id(self, client_id: str) -> str:
        """
        Retrieve `thing_id` mapped to a client. The value must have been previously set using `register` method.

        Parameters
        ----------
        client_id: str
            the client id for which the `thing_id` is to be retrieved
        """
        if client_id not in self._client_to_thing_map:
            raise ValueError(f"thing_id for client_id '{client_id}' not present in pool")
        return self._client_to_thing_map.get(client_id, None)

    def get_client_protocol(self, thing_id: str) -> str:
        """
        Retrieve protocol used by a client in the pool.

        Parameters
        ----------
        client_id: str
            the client id for which the protocol is to be retrieved
        """
        if thing_id not in self._thing_to_client_map:
            raise ValueError(f"client_id '{thing_id}' not present in pool")
        client_id = self._thing_to_client_map[thing_id]
        return self.pool[client_id].socket_address.split("://")[0].upper()

    @property
    def poll_timeout(self) -> int:
        """socket polling timeout in milliseconds greater than 0"""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, value) -> None:
        if not isinstance(value, int) or value < 0:
            raise ValueError(
                "polling period must be an integer greater than 0, not {}. Value is considered in milliseconds".format(
                    value
                )
            )
        self._poll_timeout = value

    async def handshake_complete(self) -> None:
        """wait for handshake to complete for all clients in the pool"""
        for client in self.pool.values():
            await client.handshake_complete()  # sufficient to wait serially

    def handshake(self, timeout: int | None = 60000) -> None:
        """
        schedules handshake coroutines for each client in the running event loop
        or completes handshake synchronously if no event loop is running.
        Use `handshake_complete()` async method to check if handshake is complete.

        Parameters
        ----------
        timeout: float | int
            timeout in milliseconds to wait for handshake to complete. If handshake does not complete within this time,
            a `ConnectionError` is raised. If None, wait indefinitely until handshake completes.
        """
        for client in self.pool.values():
            client.handshake(timeout)

    async def poll_responses(self) -> None:
        """
        Poll for replies from server.This method should be independently started in the event loop by calling `start_polling()`.
        Sending message requests and retrieving a response is still carried out by other methods.
        Do not duplicate this method call as there are no checks how many pollers exist and messages will become malformed
        if multiple pollers are active.
        """
        self.logger.info("client polling started for sockets for {}".format(list(self.pool.keys())))
        self.stop_poll = False
        event_loop = asyncio.get_event_loop()
        while not self.stop_poll:
            sockets = await self.poller.poll(self.poll_timeout)  # type hints dont work in this line
            for socket, _ in sockets:
                while True:
                    try:
                        raw_response = await socket.recv_multipart(zmq.NOBLOCK)
                        response_message = ResponseMessage(raw_response)
                    except zmq.Again:
                        # errors in handle_message should reach the client.
                        break
                    except ConnectionAbortedError:
                        for client in self.pool.values():
                            if client.socket.get_monitor_socket() == socket:
                                self.poller.unregister(client.socket)  # leave the monitor in the pool
                                client.handshake(timeout=None)
                                self.logger.error(
                                    f"client {client.id} disconnected from server."
                                    + " Unregistering from poller temporarily until server comes back."
                                )
                                break
                    else:
                        if self.handled_default_message_types(response_message):
                            continue
                        message_id = response_message.id
                        self.logger.debug(
                            "received response from server",
                            sender_id=response_message.sender_id,
                            receiver_id=response_message.receiver_id,
                            msg_id=response_message.id,
                            message_type=response_message.type,
                        )
                        if message_id in self.cancelled_messages:
                            self.cancelled_messages.remove(message_id)
                            self.logger.debug("dropping a cancelled message", msg_id=message_id)
                            continue
                        self.message_map[message_id] = response_message
                        event = self.events_map.get(message_id, None)
                        if event:
                            event.set()
                        else:
                            event_loop.create_task(self._resolve_response(message_id, response_message))

    async def _resolve_response(self, message_id: str, data: Any) -> None:
        """
        This method is called when there is no asyncio Event available for a message ID. This can happen only
        when the server replied before the client created a asyncio.Event object. check `async_execute()` for details.

        Parameters
        ----------
        message_id: bytes
            the message for which the event was not created
        data: ResponseMessage
            the response message received from server
        """
        max_number_of_retries = 100
        for i in range(max_number_of_retries):
            await asyncio.sleep(0.025)
            try:
                event = self.events_map[message_id]
            except KeyError:
                if message_id in self.cancelled_messages:
                    # Only for safety, likely should never reach here
                    self.cancelled_messages.remove(message_id)
                    self.logger.debug("message cancelled, not retrieving response", msg_id=message_id)
                    return
                if i >= max_number_of_retries - 1:
                    self.logger.error("unknown message id without corresponding event object", msg_id=message_id)
                    return
            else:
                self.message_map[message_id] = data
                event.set()
                break

    def assert_client_ready(self, client: AsyncZMQClient):
        if not client._handshake_event.is_set():
            raise ConnectionAbortedError(f"{client.id} is currently not alive")
        if client.socket not in self.poller._map:
            raise ConnectionError(
                "handshake complete, server is alive but client socket not yet ready to be polled."
                + "Application using MessageMappedClientPool should register the socket manually for polling."
                + "If using hololinked.server.HTTPServer, socket is waiting until HTTP Server updates its "
                "routing logic as the server has just now come alive, please try again soon."
            )

    async def async_send_request(
        self,
        thing_id: str,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> str:
        """
        send request message to server.

        Parameters
        ----------
        client_id: str
            id of the client in the pool to be used to send the message
        thing_id: str
            `id` of the `Thing` on which an operation is to be performed
        objekt: str
            name of property, action or event (usually only property or action)
        operation: str
            operation to be performed, like `readproperty`, `writeproperty`, `invokeaction` etc.
        payload: SerializableData
            serializable data to be sent as payload
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        server_execution_context: dict[str, Any]
            Specify server level execution context like `invokationTimeout`, `executionTimeout`, `oneway` operation etc.
        thing_execution_context: dict[str, Any]
            Specify thing level execution context like `fetchExecutionLogs` etc.

        Returns
        -------
        bytes
            a message id in bytes
        """
        client_id = self.get_client_id_from_thing_id(thing_id)
        self.assert_client_ready(self.pool[client_id])
        message_id = await self.pool[client_id].async_send_request(
            thing_id=thing_id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )
        event = self.event_pool.pop()
        self.events_map[message_id] = event
        return message_id

    async def async_recv_response(
        self, thing_id: str, message_id: bytes, timeout: float | int | None = None
    ) -> ResponseMessage:
        """
        Receive response for specified message ID.

        Parameters
        ----------
        client_id: str
            id of the client in the pool to be used to receive the message
        message_id: bytes
            the message id for which response needs to be fetched
        timeout: float | int | None
            Client side timeout, not the same as timeout passed to server. Recommended to be None in general cases.
            The reply is dropped if this timeout occurs. Usually the server always replies when either of invokation or
            execution timeout occurs. This timeout ensures that the protocol binding does not wait indefinitely for a message
            that may never arrive.

        Returns
        -------
        response: ResponseMessage
            response message from server corresponding to the message id

        Raises
        ------
        ValueError
            if supplied message id is not valid
        TimeoutError
            if timeout is not None and response did not arrive
        """
        try:
            client_id = self.get_client_id_from_thing_id(thing_id)
            event = self.events_map[message_id]
        except KeyError:
            raise KeyError(f"message id {message_id} unknown.") from None
        while True:
            try:
                await asyncio.wait_for(event.wait(), timeout)
                # default 5 seconds because we want to check if server is also dead
                if event.is_set():  # i.e. if timeout is not None, check if event is set
                    break
                self.assert_client_ready(self.pool[client_id])
            except TimeoutError:
                self.cancelled_messages.append(message_id)
                self.logger.error("message added to list of cancelled messages", msg_id=message_id)
                raise TimeoutError(f"Execution not completed within {timeout} seconds") from None
        self.events_map.pop(message_id)
        self.event_pool.completed(event)
        response = self.message_map.pop(message_id)
        return response

    async def async_execute(
        self,
        thing_id: str,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> ResponseMessage:
        """
        send an operation and receive the response for it.

        Parameters
        ----------
        client_id: str
            id of the client in the pool to be used to send the message
        thing_id: str
            `id` of the `Thing` on which an operation is to be performed
        objekt: str
            name of property, action or event (usually only property or action)
        operation: str
            operation to be performed, like `readproperty`, `writeproperty`, `invokeaction` etc.
        payload: SerializableData
            serializable data to be sent as payload
        preserialized_payload: PreserializedData
            pre-encoded data to be sent as payload, generally used for large or custom data that is already serialized
        server_execution_context: dict[str, Any]
            Specify server level execution context like `invokationTimeout`, `executionTimeout`, `oneway` operation etc.
        thing_execution_context: dict[str, Any]
            Specify thing level execution context like `fetchExecutionLogs` etc.

        Returns
        -------
        ResponseMessage
            response message from server after completing the operation
        """
        message_id = await self.async_send_request(
            thing_id=thing_id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )
        return await self.async_recv_response(
            thing_id=thing_id,
            message_id=message_id,
        )

    def start_polling(self) -> None:
        """register the server message polling loop in the asyncio event loop"""
        get_current_async_loop().create_task(self.poll_responses())

    def stop_polling(self):
        """
        stop polling for replies from server
        """
        self.stop_poll = True
        for client in self.pool.values():
            client.stop()

    async def async_execute_in_all(
        self,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        thing_ids: list[str] | None = None,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> dict[str, ResponseMessage]:
        if not thing_ids:
            thing_ids = self._client_to_thing_map.values()

        gathered_replies = await asyncio.gather(
            *[
                self.async_execute(
                    thing_id=id,
                    objekt=objekt,
                    operation=operation,
                    payload=payload,
                    preserialized_payload=preserialized_payload,
                    server_execution_context=server_execution_context,
                    thing_execution_context=thing_execution_context,
                )
                for id in thing_ids
            ]
        )
        replies = dict()
        for id, response in zip(thing_ids, gathered_replies):
            replies[id] = response
        return replies

    async def async_execute_in_all_things(
        self,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> dict[str, ResponseMessage]:
        """execute the same operation in all `Thing`s"""
        return await self.async_execute_in_all(
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )

    async def ping_all_servers(self):
        return await self.async_execute_in_all()  # operation='invokeAction', objekt=CommonRPC.PING)

    def __contains__(self, name: str) -> bool:
        return name in self.pool

    def __getitem__(self, key) -> AsyncZMQClient:
        return self.pool[key]

    def __iter__(self) -> Iterator[AsyncZMQClient]:
        return iter(self.pool.values())

    def exit(self) -> None:
        try:
            BaseZMQ.exit(self)
            for client in self.pool.values():
                self.poller.unregister(client.socket)
                self.poller.unregister(client.socket.get_monitor_socket())
                client.exit()
            self.logger.info("all client socket unregistered")
        except Exception as ex:
            self.logger.warning(
                "could not properly terminate context or attempted to terminate an already terminated context."
                + f" Exception message: {str(ex)}"
            )

    """
    Inheritance tree:

    BaseZMQ
    BaseAsyncZMQ
    BaseSyncZMQ
    BaseZMQClient
    SyncZMQClient
    AsyncZMQClient
    MessageMappedClientPool
    """


class AsyncioEventPool:
    """
    creates a pool of asyncio Events to be used as a synchronisation object for MessageMappedClientPool

    Parameters
    ----------
    initial_number_of_events: int
        initial pool size of events
    """

    def __init__(self, initial_number_of_events: int) -> None:
        self.pool = [asyncio.Event() for i in range(initial_number_of_events)]
        self.size = initial_number_of_events

    def pop(self) -> asyncio.Event:
        """
        pop an event, new one is created if nothing left in pool
        """
        try:
            event = self.pool.pop(0)
        except IndexError:
            self.size += 1
            event = asyncio.Event()
        event.clear()
        return event

    def completed(self, event: asyncio.Event) -> None:
        """
        put an event back into the pool
        """
        self.pool.append(event)


class EventPublisher(BaseZMQServer, BaseSyncZMQ):
    """Event publisher for broadcasting messages to all connected clients. Implements PUB-SUB pattern."""

    _standard_address_suffix = "/event-publisher"
    _standard_address_suffix_filename_replacement = "event-publisher"

    def __init__(
        self,
        id: str,
        context: zmq.Context | None = None,
        access_point: str = ZMQ_TRANSPORTS.IPC,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            unique identifier of the publisher
        context: zmq.Context
            ZMQ context to use, if None, a global context is used.
        access_point: Enum | str, default ZMQ_TRANSPORTS.IPC
            access point for the publisher to bind to, usually `INPROC`
        """
        super().__init__(id=id, **kwargs)
        self.create_socket(
            server_id=id,
            socket_id=id,
            node_type="server",
            context=context,
            access_point=access_point,
            socket_type=zmq.PUB,
            **kwargs,
        )
        self.events = set()  # type is set[EventDispatcher]
        self.event_ids = set()  # type: set[str]
        self._send_lock = threading.Lock()

    def register(self, event: "EventDispatcher") -> None:
        """
        register event with a specific (unique) name

        Parameters
        ----------
        event: `EventDispatcher`
            `Event` object that needs to be registered. Events created at `__init__()` of `Thing` are
            automatically registered.
        """
        if event._unique_identifier in self.events and event not in self.events:
            raise AttributeError(f"event {event._unique_identifier} already registered, please use another name.")
        self.event_ids.add(event._unique_identifier)
        self.events.add(event)

    def unregister(self, event: "EventDispatcher") -> None:
        """
        unregister event with a specific (unique) name

        Parameters
        ----------
        event: `EventDispatcher`
            `Event` object that needs to be unregistered.
        """
        if event in self.events:
            self.events.remove(event)
            self.event_ids.remove(event._unique_identifier)
        else:
            warnings.warn(
                f"event {event._unique_identifier} not found, did you mean to unregister another event?",
                UserWarning,
            )

    def publish(self, event, data: Any) -> None:
        """
        publish an event with given unique name.

        Parameters
        ----------
        event: `EventDispatcher`
            `Event` object that needs to be published.
        data: Any
            data to be sent as payload of the event
        """
        # uncomment for type definitions
        # from ...core.events import EventDispatcher
        # assert isinstance(event, EventDispatcher), "event must be an instance of EventDispatcher"

        try:
            self._send_lock.acquire()
            if event._unique_identifier in self.event_ids:
                serializer = Serializers.for_object(
                    event._owner_inst.id,
                    event._owner_inst.__class__.__name__,
                    event._descriptor.name,
                )
                content_type_if_no_serializer = Serializers.get_content_type_for_object(
                    event._owner_inst.id,
                    event._owner_inst.__class__.__name__,
                    event._descriptor.name,
                )
                if not isinstance(data, bytes):
                    payload = SerializableData(data, serializer=serializer)
                    preserialized_payload = PreserializedEmptyByte
                else:
                    payload = SerializableNone
                    preserialized_payload = PreserializedData(data, content_type=content_type_if_no_serializer)

                event_message = EventMessage.craft_from_arguments(
                    event._unique_identifier,
                    self.id,
                    payload=payload,
                    preserialized_payload=preserialized_payload,
                )
                self.socket.send_multipart(event_message.byte_array)
                self.logger.debug(f"published event with unique identifier {event._unique_identifier}")
                # print("published event with unique identifier {}".format(event._unique_identifier))
                return
            raise AttributeError(f"event name {event._unique_identifier} not registered")
        finally:
            try:
                self._send_lock.release()
            except Exception as ex:
                self.logger.warning(f"could not release publish lock for event publisher - {str(ex)}")

    def exit(self):
        try:
            BaseZMQ.exit(self)
            self.socket.close(0)
            self.logger.info("terminated event publishing socket")
        except Exception as ex:
            self.logger.warning(
                "could not properly terminate context or attempted to terminate an already terminated context."
                + f" Exception message: {str(ex)}"
            )


class BaseEventConsumer(BaseZMQClient):
    """Consumes events published at PUB sockets using SUB socket"""

    def __init__(
        self,
        id: str,
        event_unique_identifier: str,
        access_point: str,
        context: zmq.Context | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            unique identity for the consumer
        event_unique_identifier: str
            unique identifier of the event registered at the PUB socket
        access_point: str
            socket address of the event publisher (`EventPublisher`), properly qualified with transport method
        context: zmq.Context
            ZMQ context to use, if None, a global context is used.
        **kwargs:
            additional arguments:

            - `logger`: `logging.Logger`, logger instance to use. If None, a default
            - `poll_timeout`: `int`, socket polling timeout in milliseconds greater than 0.
            - `server_id`: `str`, id of the PUB socket server, usually not necessary as `access_point` is sufficient.
        """

        if isinstance(self, BaseSyncZMQ):
            self.context = context or global_config.zmq_context()
            self.poller = zmq.Poller()
            socket_class = zmq.Socket
            self._poller_lock = threading.Lock()
        elif isinstance(self, BaseAsyncZMQ):
            self.context = context or global_config.zmq_context()
            self.poller = zmq.asyncio.Poller()
            socket_class = zmq.asyncio.Socket
            self._poller_lock = asyncio.Lock()
        else:
            raise TypeError("BaseEventConsumer must be subclassed by either BaseSyncZMQ or BaseAsyncZMQ")
        super().__init__(id=id, server_id=kwargs.get("server_id", None), **kwargs)
        logger = kwargs.get("logger", None)
        if not logger:
            logger = structlog.get_logger().bind(
                component="broker",
                impl=self.__class__.__name__,
                id=id,
                event_id=event_unique_identifier,
            )
        self.logger = logger  # type: structlog.stdlib.BoundLogger
        self.create_socket(
            server_id=id,
            socket_id=id,
            node_type="client",
            context=self.context,
            socket_type=zmq.SUB,
            access_point=access_point,
            **kwargs,
        )
        self.event_unique_identifier = bytes(event_unique_identifier, encoding="utf-8")
        short_uuid = uuid_hex()
        self.interruptor = self.context.socket(zmq.PAIR, socket_class=socket_class)
        self.interruptor.setsockopt_string(zmq.IDENTITY, f"interrupting-server-{short_uuid}")
        self.interrupting_peer = self.context.socket(zmq.PAIR, socket_class=socket_class)
        self.interrupting_peer.setsockopt_string(zmq.IDENTITY, f"interrupting-client-{short_uuid}")
        self.interruptor.bind(f"inproc://{self.id}-{short_uuid}/interruption")
        self.interrupting_peer.connect(f"inproc://{self.id}-{short_uuid}/interruption")
        self._stop = False

    def subscribe(self) -> None:
        """subscribe to the event at the PUB socket"""
        self.socket.setsockopt(zmq.SUBSCRIBE, self.event_unique_identifier)
        # pair sockets cannot be polled unforunately, so we use router
        # if self.socket in self.poller._map:
        #     self.poller.unregister(self.socket)
        # if self.interruptor in self.poller._map:
        #     self.poller.unregister(self.interruptor)
        self.poller.register(self.socket, zmq.POLLIN)
        self.poller.register(self.interruptor, zmq.POLLIN)

    def stop_polling(self) -> None:
        """stop polling for events when `receive()` is called"""
        self._stop = True

    @property
    def interrupt_message(self) -> EventMessage:
        """
        craft an interrupt message to be sent to the interruptor socket, if `stop_polling()` is not sufficient as
        the poll timeout is infinite. Used internally by `interrupt()` method.
        """
        return EventMessage.craft_from_arguments(
            event_id=f"{self.id}/interrupting-server",
            sender_id=self.id,
            payload=SerializableData("INTERRUPT", content_type="application/json"),
        )

    def exit(self):
        try:
            BaseZMQ.exit(self)
            self.poller.unregister(self.socket)
            self.poller.unregister(self.interruptor)
        except Exception as ex:  # noqa
            # TODO - log message and undo noqa
            self.logger.warning(f"could not unregister sockets from poller for event consumer - {str(ex)}")
        try:
            self.socket.close(0)
            self.interruptor.close(0)
            self.interrupting_peer.close(0)
            self.logger.info(f"terminated event consuming socket {self.socket_address}")
        except Exception as ex:
            self.logger.warning(f"could not terminate sockets. exception message - {str(ex)}")


class EventConsumer(BaseEventConsumer, BaseSyncZMQ):
    """Sync Event Consumer to used outside of async loops"""

    def receive(self, timeout: float | None = 1000, raise_interrupt_as_exception: bool = False) -> EventMessage | None:
        """
        receive event with given timeout

        Parameters
        ----------
        timeout: float, int, None
            timeout in milliseconds, None for blocking
        raise_interrupt_as_exception: bool
            if True, raises BreakLoop exception when interrupted, otherwise returns None
        """
        self._stop = False
        while not self._stop:
            try:
                if not self._poller_lock.acquire(timeout=timeout / 1000 if timeout else -1):
                    continue
                sockets = self.poller.poll(timeout)  # list[tuple[zmq.Socket, int]]
                if len(sockets) > 1:
                    # if there is an interrupt message as well as an event,
                    # give preference to interrupt message.
                    if sockets[0][0] == self.interrupting_peer:
                        sockets = [sockets[0]]  # we still need the socket, poll event  tuple
                    elif sockets[1][0] == self.interrupting_peer:
                        sockets = [sockets[1]]
                for socket, _ in sockets:
                    try:
                        raw_message = socket.recv_multipart(zmq.NOBLOCK)
                        message = EventMessage(raw_message)
                        if socket == self.interrupting_peer:
                            if message.payload.deserialize() == "INTERRUPT":
                                self.stop_polling()
                                if raise_interrupt_as_exception:
                                    raise BreakLoop("event consumer interrupted")
                                return
                        return message
                    except zmq.Again:
                        pass
                    # if not self.handled_default_message_types(event_message):
            finally:
                try:
                    self._poller_lock.release()
                except Exception as ex:
                    self.logger.warning(f"could not release poller lock for event receive - {str(ex)}")

    def interrupt(self):
        """
        interrupts the event consumer. Generally should be used for exiting this object if there is no poll
        period/infinite polling. Otherwise please use stop_polling().
        """
        self.interrupting_peer.send_multipart(self.interrupt_message.byte_array)


class AsyncEventConsumer(BaseEventConsumer, BaseAsyncZMQ):
    """Async Event Consumer to be used inside async loops"""

    async def receive(
        self,
        timeout: float | None = 1000,
        raise_interrupt_as_exception: bool = False,
    ) -> EventMessage | None:
        """
        receive event with given timeout

        Parameters
        ----------
        timeout: float, int, None
            timeout in milliseconds, None for blocking
        raise_interrupt_as_exception: bool
            if True, raises BreakLoop exception when interrupted, otherwise returns None
        """
        # TODO - use raise_interrupt_as_exception
        self._stop = False
        while not self._stop:
            try:
                try:
                    await asyncio.wait_for(
                        self._poller_lock.acquire(),
                        timeout=timeout / 1000 if timeout else None,
                    )
                except TimeoutError:
                    continue
                sockets = await self.poller.poll(timeout)
                if len(sockets) > 1:
                    # if there is an interrupt message as well as an event,
                    # give preference to interrupt message.
                    if sockets[0][0] == self.interrupting_peer:
                        sockets = [sockets[0]]
                    elif sockets[1][0] == self.interrupting_peer:
                        sockets = [sockets[1]]
                for socket, _ in sockets:
                    try:
                        raw_message = await socket.recv_multipart(zmq.NOBLOCK)
                        message = EventMessage(raw_message)
                        if socket == self.interrupting_peer:
                            if message.payload.deserialize() == "INTERRUPT":
                                self.stop_polling()
                                if raise_interrupt_as_exception:
                                    raise BreakLoop("event consumer interrupted")
                                return
                        return message
                    except zmq.Again:
                        pass
            finally:
                try:
                    self._poller_lock.release()
                except Exception as ex:
                    self.logger.warning(f"could not release poller lock for event receive - {str(ex)}")

    async def interrupt(self):
        """
        interrupts the event consumer. Generally should be used for exiting this object if there is no poll
        period/infinite polling. Otherwise please use stop_polling().
        """
        await self.interrupting_peer.send_multipart(self.interrupt_message.byte_array)


from ...core.events import EventDispatcher  # noqa

__all__ = [
    AsyncZMQServer.__name__,
    ZMQServerPool.__name__,
    SyncZMQClient.__name__,
    AsyncZMQClient.__name__,
    MessageMappedZMQClientPool.__name__,
    AsyncEventConsumer.__name__,
    EventConsumer.__name__,
]



from typing import Any

import pytest
import structlog

from things import OceanOpticsSpectrometer

from hololinked.core import Action, Event, Property, Thing, ThingMeta
from hololinked.core.actions import BoundAction
from hololinked.core.events import EventDispatcher
from hololinked.core.logger import RemoteAccessHandler
from hololinked.core.meta import (
    ActionsRegistry,
    DescriptorRegistry,  # noqa: F401
    EventsRegistry,
    PropertiesRegistry,
)
from hololinked.core.properties import Parameter  # noqa: F401
from hololinked.core.state_machine import BoundFSM
from hololinked.core.zmq.brokers import EventPublisher
from hololinked.core.zmq.rpc_server import RPCServer


"""
The tests in this file are for the initialization of the Thing class and its subclasses.
1. Test Thing class 
2. Test Thing subclass
3. Test ThingMeta metaclass
4. Test ActionRegistry class
5. Test EventRegistry class
6. Test PropertiesRegistry class

Test sequence is as follows:
1. Test id requirements 
2. Test logger setup
3. Test state and state_machine setup
4. Test composition of subthings
5. Test servers init
6. Test thing model generation
"""


@pytest.mark.parametrize("thing_cls", [Thing, OceanOpticsSpectrometer])
def test_01_id(thing_cls: ThingMeta):
    """Test id property of Thing class"""
    # req. 1. instance name must be a string and cannot be changed after set
    thing = thing_cls(id="test_id")  # type: Thing
    assert thing.id == "test_id"
    with pytest.raises(ValueError):
        thing.id = "new_instance"
    with pytest.raises(NotImplementedError):
        del thing.id
    # req. 2. regex is r'[A-Za-z]+[A-Za-z_0-9\-\/]*', simple URI like
    valid_ids = ["test_id", "A123", "valid_id-123", "another/valid-id"]
    invalid_ids = ["123_invalid", "invalid id", "invalid@id", ""]
    for valid_id in valid_ids:
        thing.properties.descriptors["id"].validate_and_adapt(valid_id)
    for invalid_id in invalid_ids:
        with pytest.raises(ValueError):
            thing.properties.descriptors["id"].validate_and_adapt(invalid_id)


@pytest.mark.parametrize("thing_cls", [Thing, OceanOpticsSpectrometer])
def notest_02_logger(thing_cls: ThingMeta):
    """Test logger setup"""
    # req. 1. logger must have remote access handler if remote_accessible_logger is True
    logger = structlog.get_logger("test_logger")
    thing = thing_cls(
        id="test_remote_accessible_logger",
        logger=logger,
        remote_accessible_logger=True,
    )  # type: Thing
    assert thing.logger == logger
    assert any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers)
    # Therefore also check the false condition
    logger = structlog.get_logger("test_logger_no_remote_access")
    thing = thing_cls(
        id="test_logger_without_remote_access",
        logger=logger,
        remote_accessible_logger=False,
    )  # type: Thing
    assert not any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers)
    # NOTE - logger is modifiable after instantiation

    # req. 2. logger is created automatically if not provided
    thing = thing_cls(id="test_logger_auto_creation")
    assert thing.logger is not None
    assert not any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers)
    assert thing.logger != logger  # not the above logger that we used.
    # remote accessible only when we ask for it
    thing = thing_cls(id="test_logger_auto_creation_2", remote_accessible_logger=True)  # type: Thing
    assert thing.logger is not None
    assert any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers)
    assert thing.logger != logger


@pytest.mark.parametrize("thing_cls", [Thing])
def test_03_has_no_fsm(thing_cls: ThingMeta):
    """Test state and state_machine setup"""
    # req. 1. state property must be None when no state machine is present
    thing = thing_cls(id="test_no_state_machine")  # type: Thing
    if thing.state_machine is None:
        assert thing.state is None
        assert thing.state_machine is None


@pytest.mark.parametrize("thing_cls", [OceanOpticsSpectrometer])
def test_04_bound_fsm(thing_cls: ThingMeta):
    """Test state and state_machine setup"""
    thing1 = thing_cls(id="test_state_machine")  # type: Thing
    # req. 1. state and state machine must be present because we create this subclass with a state machine
    assert thing1.state is not None
    assert isinstance(thing1.state_machine, BoundFSM)
    # req. 2. state and state machine must be different for different instances
    thing2 = thing_cls(id="test_state_machine_2")  # type: Thing
    # first check if state machine exists
    assert thing2.state is not None
    assert isinstance(thing2.state_machine, BoundFSM)
    # then check if they are different
    assert thing1.state_machine != thing2.state_machine
    # until state is set, initial state is equal
    assert thing1.state == thing2.state
    assert thing1.state_machine.initial_state == thing2.state_machine.initial_state
    # after state is set, they are different
    thing1.state_machine.set_state(thing1.states.ALARM)
    assert thing1.state != thing2.state
    assert thing1.state_machine != thing2.state_machine
    # initial state is still same
    assert thing1.state_machine.initial_state == thing2.state_machine.initial_state
    # detailed checks in another file


@pytest.mark.parametrize("thing_cls", [Thing, OceanOpticsSpectrometer])
def test_05_subthings(thing_cls: ThingMeta):
    """Test object composition"""
    thing = thing_cls(id="test_subthings", remote_accessible_logger=True)  # type: Thing
    # req. 1. subthings must be a dictionary
    assert isinstance(thing.sub_things, dict)
    assert len(thing.sub_things) == 1  # logger
    # req. 2. subthings are always recomputed when accessed (at least thats the way it is right now),
    # so we can add new subthings anytime
    thing.another_thing = OceanOpticsSpectrometer(id="another_thing")
    assert isinstance(thing.sub_things, dict)
    assert len(thing.sub_things) == 2  # logger + another_thing
    # req. 3. subthings must be instances of Thing and have the parent as owner
    for name, subthing in thing.sub_things.items():
        assert thing in subthing._owners  # type: ignore[attr-defined]
        assert isinstance(subthing, Thing)
        # req. 4. name of subthing must match name of the attribute
        assert hasattr(thing, name)


@pytest.mark.parametrize("thing_cls", [Thing, OceanOpticsSpectrometer])
def test_06_servers_init(thing_cls: ThingMeta):
    """Test if servers can be initialized/instantiated"""
    # req. 1. rpc_server and event_publisher must be None when not run()
    thing = thing_cls(id="test_servers_init")  # type: Thing
    assert thing.rpc_server is None
    assert thing.event_publisher is None
    # req. 2. rpc_server and event_publisher must be instances of their respective classes when run()
    RPCServer(id="test-rpc-server-init", things=[thing], logger=thing.logger)  # prepare server class
    assert isinstance(thing.rpc_server, RPCServer)
    assert isinstance(thing.event_publisher, EventPublisher)
    # exit to quit nicely
    thing.rpc_server.exit()
    thing.event_publisher.exit()


"""
Test sequence is as follows:
1. Test metaclass of Thing class
2. Test registry creation and access which is currently the main purpose of the metaclass
"""


@pytest.mark.parametrize("thing_cls", [Thing, OceanOpticsSpectrometer])
def test_07_metaclass_assigned(thing_cls: ThingMeta):
    """test metaclass of Thing class"""
    # req. 1 metaclass must be ThingMeta of any Thing class
    assert thing_cls.__class__ == ThingMeta
    assert OceanOpticsSpectrometer.__class__ == ThingMeta
    assert Thing.__class__ == OceanOpticsSpectrometer.__class__


def test_08_registry_creation():
    """test registry creation and access which is currently the main purpose of the metaclass"""
    # req. 1. registry attributes must be instances of their respective classes
    assert isinstance(Thing.properties, PropertiesRegistry)
    assert isinstance(Thing.actions, ActionsRegistry)
    assert isinstance(Thing.events, EventsRegistry)

    # req. 2. new registries are not created on the fly and are same between accesses
    assert Thing.properties == Thing.properties
    assert Thing.actions == Thing.actions
    assert Thing.events == Thing.events
    # This test is done as the implementation deviates from `param`

    # req. 3. different subclasses have different registries
    assert Thing.properties != OceanOpticsSpectrometer.properties
    assert Thing.actions != OceanOpticsSpectrometer.actions
    assert Thing.events != OceanOpticsSpectrometer.events

    # create instances for further tests
    thing = Thing(id="test_registry_creation")
    spectrometer = OceanOpticsSpectrometer(id="test_registry_creation_2")

    # req. 4. registry attributes must be instances of their respective classes also for instances
    assert isinstance(thing.properties, PropertiesRegistry)
    assert isinstance(thing.actions, ActionsRegistry)
    assert isinstance(thing.events, EventsRegistry)

    # req. 5. registries are not created on the fly and are same between accesses also for instances
    assert thing.properties == thing.properties
    assert thing.actions == thing.actions
    assert thing.events == thing.events

    # req. 6. registries are not shared between instances
    assert thing.properties != spectrometer.properties
    assert thing.actions != spectrometer.actions
    assert thing.events != spectrometer.events

    # req. 7. registries are not shared between instances and their classes
    assert thing.properties != Thing.properties
    assert thing.actions != Thing.actions
    assert thing.events != Thing.events
    assert spectrometer.properties != OceanOpticsSpectrometer.properties
    assert spectrometer.actions != OceanOpticsSpectrometer.actions
    assert spectrometer.events != OceanOpticsSpectrometer.events


"""
Test action registry first because actions are the easiest to test.
1. Test owner attribute
2. Test descriptors access
3. Test dunders
"""


class Registry:
    """Class to hold registry class and object for parameterized tests"""

    cls: type[PropertiesRegistry | ActionsRegistry | EventsRegistry]
    cls_object: PropertiesRegistry | ActionsRegistry | EventsRegistry
    inst_object: PropertiesRegistry | ActionsRegistry | EventsRegistry | None
    obj: type[Property | Action | Event]
    bound_object: type[BoundAction | EventDispatcher] | Any  # any is for property value
    thing_cls: ThingMeta
    thing_inst: Thing

    def __init__(self) -> None:
        pass


@pytest.fixture(
    params=[
        pytest.param((Thing, PropertiesRegistry), id="Thing-PropertiesRegistry"),
        pytest.param((Thing, ActionsRegistry), id="Thing-ActionsRegistry"),
        pytest.param((Thing, EventsRegistry), id="Thing-EventsRegistry"),
        pytest.param((OceanOpticsSpectrometer, PropertiesRegistry), id="OceanOpticsSpectrometer-PropertiesRegistry"),
        pytest.param((OceanOpticsSpectrometer, ActionsRegistry), id="OceanOpticsSpectrometer-ActionsRegistry"),
        pytest.param((OceanOpticsSpectrometer, EventsRegistry), id="OceanOpticsSpectrometer-EventsRegistry"),
    ],
)
def registry(request) -> Registry:
    # create instances for further tests
    cls, registry_cls = request.param
    thing = cls(id=f"test_{registry_cls.__name__}_registry")
    registry = Registry()
    registry.thing_cls = cls
    registry.thing_inst = thing
    registry.cls = registry_cls
    if registry_cls == ActionsRegistry:
        registry.cls_object = cls.actions
        registry.inst_object = thing.actions
        registry.obj = Action
        registry.bound_object = BoundAction
    elif registry_cls == PropertiesRegistry:
        registry.cls_object = cls.properties
        registry.inst_object = thing.properties
        registry.obj = Parameter
        registry.bound_object = Any
    elif registry_cls == EventsRegistry:
        registry.cls_object = cls.events
        registry.inst_object = thing.events
        registry.obj = Event
        registry.bound_object = EventDispatcher
    else:
        raise NotImplementedError("This registry class is not implemented")
    return registry


def test_09_registry_owner(registry: Registry):
    """Test owner attribute of DescriptorRegistry"""
    # See comment above TestRegistry class to enable type definitions
    # req. 1. owner attribute must be the class itself when accessed as class attribute
    assert registry.cls_object.owner == registry.thing_cls
    # therefore owner instance must be None
    assert registry.cls_object.owner_inst is None

    # req. 2. owner attribute must be the instance for instance registries (i.e. when accessed as instance attribute)
    assert registry.inst_object.owner == registry.thing_inst
    assert registry.inst_object.owner_cls == registry.thing_cls

    # req. 3. descriptor_object must be defined correctly and is a class
    assert registry.cls_object.descriptor_object == registry.obj
    assert registry.inst_object.descriptor_object == registry.obj
    assert registry.cls_object.descriptor_object == registry.inst_object.descriptor_object


def test_10_descriptors_access(registry: Registry):
    """Test descriptors access"""

    # req. 1. descriptors are instances of the descriptor object - Property | Action | Event
    for name, value in registry.cls_object.descriptors.items():
        assert isinstance(value, registry.obj)
        assert isinstance(name, str)

    # req. 2. either class level or instance level descriptors are same - not a strict requirement for different
    # use cases, one can always add instance level descriptors
    for name, value in registry.inst_object.descriptors.items():
        assert isinstance(value, registry.obj)
        assert isinstance(name, str)

    # req. 3. because class level and instance level descriptors are same, they are equal
    for (name, value), (name2, value2) in zip(
        registry.cls_object.descriptors.items(),
        registry.inst_object.descriptors.items(),
    ):
        assert name == name2
        assert value == value2

    # req. 4. descriptors can be cleared
    assert hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{registry.cls.__name__.lower()}",
    )
    registry.inst_object.clear()
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{registry.cls.__name__.lower()}",
    )
    # clearing again any number of times should not raise error
    registry.inst_object.clear()
    registry.inst_object.clear()
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{registry.cls.__name__.lower()}",
    )


def test_11_registry_dunders(registry: Registry):
    """Test dunders of DescriptorRegistry"""

    # req. 1. __getitem__ must return the descriptor object
    for name, value in registry.cls_object.descriptors.items():
        assert registry.cls_object[name] == value
        # req. 2. __contains__ must return True if the descriptor is present
        assert value in registry.cls_object
        assert name in registry.cls_object.descriptors.keys()

    # req. 2. __iter__ must return an iterator over the descriptors dictionary
    # which in turn iterates over the keys
    assert all(isinstance(descriptor_name, str) for descriptor_name in registry.cls_object)
    assert all(isinstance(descriptor_name, str) for descriptor_name in registry.inst_object)
    # __iter__ can also be casted as other iterators like lists
    descriptors = list(registry.inst_object)
    assert isinstance(descriptors, list)
    assert all(isinstance(descriptor_name, str) for descriptor_name in descriptors)

    # req. 3. __len__ must return the number of descriptors
    assert len(registry.cls_object) == len(registry.cls_object.descriptors)
    assert len(registry.inst_object) == len(registry.inst_object.descriptors)
    assert len(registry.inst_object) == len(registry.cls_object)

    # req. 4. registries have their unique hashes
    # NOTE - not sure if this is really a useful feature or just plain stupid
    # The requirement was to be able to generate unique hashes for each registry like foodict[<some hash>] = Thing.actions
    foodict = {
        registry.cls_object: 1,
        registry.inst_object: 3,
    }
    assert foodict[registry.cls_object] == 1
    assert foodict[registry.inst_object] == 3

    # __dir__ not yet tested
    # __str__ will not be tested


def test_12_bound_objects(registry: Registry):
    """Test bound objects returned from descriptor access"""
    # req. 1. number of bound objects must be equal to number of descriptors
    # for example, number of bound actions must be equal to number of actions
    assert len(registry.inst_object) == len(registry.inst_object.descriptors)

    # req. 2. bound objects must be instances of bound instances
    for name, value in registry.inst_object.values.items():
        if registry.bound_object != Any:
            assert isinstance(value, registry.bound_object)
        assert isinstance(name, str)


@pytest.fixture(
    params=[
        pytest.param((Thing, EventsRegistry), id="Thing-EventsRegistry"),
        pytest.param((OceanOpticsSpectrometer, EventsRegistry), id="OceanOpticsSpectrometer-EventsRegistry"),
    ],
)
def event_registry(request) -> Registry:
    cls, registry_cls = request.param
    thing = cls(id=f"test_{registry_cls.__name__}_registry")
    registry = Registry()
    registry.thing_cls = cls
    registry.thing_inst = thing
    registry.cls = registry_cls
    registry.cls_object = cls.events
    registry.inst_object = thing.events
    registry.obj = Event
    registry.bound_object = EventDispatcher
    return registry


def test_13_descriptors_access_events(event_registry: Registry):
    registry = event_registry
    # req. 5. observables and change events are also descriptors
    for name, value in registry.inst_object.observables.items():
        assert isinstance(value, Property)
        assert isinstance(name, str)
    for name, value in registry.inst_object.change_events.items():
        assert isinstance(value, Event)
        assert isinstance(name, str)
    # req. 4. descriptors can be cleared
    assert hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}",
    )
    assert hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}_change_events",
    )
    assert hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}_observables",
    )
    registry.inst_object.clear()
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}",
    )
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}_change_events",
    )
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}_observables",
    )
    registry.inst_object.clear()
    registry.inst_object.clear()
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}",
    )
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}_change_events",
    )
    assert not hasattr(
        registry.inst_object,
        f"_{registry.inst_object._qualified_prefix}_{EventsRegistry.__name__.lower()}_observables",
    )


@pytest.fixture(
    params=[
        pytest.param((Thing, PropertiesRegistry), id="Thing-PropertiesRegistry"),
        pytest.param((OceanOpticsSpectrometer, PropertiesRegistry), id="OceanOpticsSpectrometer-PropertiesRegistry"),
    ],
)
def properties_registry(request) -> Registry:
    cls, registry_cls = request.param
    thing = cls(id=f"test_{registry_cls.__name__}_registry")
    registry = Registry()
    registry.thing_cls = cls
    registry.thing_inst = thing
    registry.cls = registry_cls
    registry.cls_object = cls.properties
    registry.inst_object = thing.properties
    registry.obj = Property
    registry.bound_object = Any
    return registry


def test_14_descriptors_access_properties(properties_registry: Registry):
    registry = properties_registry

    # req. 5. parameters that are subclass of Property are usually remote objects
    for name, value in registry.thing_inst.properties.remote_objects.items():
        assert isinstance(value, Property)
        assert isinstance(name, str)

    # req. 6. db_objects, db_init_objects, db_persisting_objects, db_commit_objects are also descriptors
    for name, value in registry.thing_inst.properties.db_objects.items():
        assert isinstance(value, Property)
        assert isinstance(name, str)
        assert value.db_init or value.db_persist or value.db_commit
    for name, value in registry.thing_inst.properties.db_init_objects.items():
        assert isinstance(value, Property)
        assert isinstance(name, str)
        assert value.db_init or value.db_persist
        assert not value.db_commit
    for name, value in registry.thing_inst.properties.db_commit_objects.items():
        assert isinstance(value, Property)
        assert isinstance(name, str)
        assert value.db_commit or value.db_persist
        assert not value.db_init
    for name, value in registry.thing_inst.properties.db_persisting_objects.items():
        assert isinstance(value, Property)
        assert isinstance(name, str)
        assert value.db_persist
        assert not value.db_init  # in user given cases, this could be true, this is not strict requirement
        assert not value.db_commit  # in user given cases, this could be true, this is not strict requirement

    # req. 4. descriptors can be cleared
    assert hasattr(
        registry.thing_inst.properties,
        f"_{registry.thing_inst.properties._qualified_prefix}_{PropertiesRegistry.__name__.lower()}",
    )

    registry.thing_inst.properties.clear()
    assert not hasattr(
        registry.thing_inst.properties,
        f"_{registry.thing_inst.properties._qualified_prefix}_{PropertiesRegistry.__name__.lower()}",
    )

    registry.thing_inst.properties.clear()
    registry.thing_inst.properties.clear()
    assert not hasattr(
        registry.thing_inst.properties,
        f"_{registry.thing_inst.properties._qualified_prefix}_{PropertiesRegistry.__name__.lower()}",
    )


@pytest.fixture(
    params=[
        pytest.param((OceanOpticsSpectrometer, PropertiesRegistry), id="OceanOpticsSpectrometer-PropertiesRegistry"),
    ],
)
def spectrometer_registry(request) -> Registry:
    cls, registry_cls = request.param
    thing = cls(id=f"test_{registry_cls.__name__}_registry")
    registry = Registry()
    registry.thing_cls = cls
    registry.thing_inst = thing
    registry.cls = registry_cls
    registry.cls_object = cls.properties
    registry.inst_object = thing.properties
    registry.obj = Property
    registry.bound_object = Any
    return registry


def test_15_bulk_read_write_properties(spectrometer_registry: Registry):
    """Test bulk read and write operations for properties"""
    registry = spectrometer_registry

    # req. 1. test read in bulk for readAllProperties
    prop_values = registry.thing_inst.properties.get()
    # read value is a dictionary
    assert isinstance(prop_values, dict)
    assert len(prop_values) > 0
    # all properties are read at instance level and get only reads remote objects
    assert len(prop_values) == len(registry.thing_inst.properties.remote_objects)
    # read values are not descriptors themselves
    for name, value in prop_values.items():
        assert isinstance(name, str)
        assert not isinstance(value, Parameter)  # descriptor has been read

    # req. 2. properties can be read with new names
    prop_values = registry.thing_inst.properties.get(
        integration_time="integrationTime",
        state="State",
        trigger_mode="triggerMode",
    )
    assert isinstance(prop_values, dict)
    assert len(prop_values) == 3
    for name, value in prop_values.items():
        assert isinstance(name, str)
        assert name in ["integrationTime", "triggerMode", "State"]
        assert not isinstance(value, Parameter)

    # req. 3. read in bulk for readMultipleProperties
    prop_values = registry.thing_inst.properties.get(
        names=["integration_time", "trigger_mode", "state", "last_intensity"]
    )
    # read value is a dictionary
    assert isinstance(prop_values, dict)
    assert len(prop_values) == 4
    # read values are not descriptors themselves
    for name, value in prop_values.items():
        assert isinstance(name, str)
        assert name in ["integration_time", "trigger_mode", "state", "last_intensity"]
        assert not isinstance(value, Parameter)

    # req. 4. read a property that is not present raises AttributeError
    with pytest.raises(AttributeError) as ex:
        prop_values = registry.thing_inst.properties.get(
            names=[
                "integration_time",
                "trigger_mode",
                "non_existent_property",
                "last_intensity",
            ]
        )
    assert "property non_existent_property does not exist" in str(ex.value)

    # req. 5. write in bulk
    prop_values = registry.thing_inst.properties.get()
    registry.thing_inst.properties.set(integration_time=10, trigger_mode=1)
    assert prop_values["integration_time"] != registry.thing_inst.integration_time
    assert prop_values["trigger_mode"] != registry.thing_inst.trigger_mode

    # req. 6. writing a non existent property raises RuntimeError
    with pytest.raises(RuntimeError) as ex:
        registry.thing_inst.properties.set(integration_time=120, trigger_mode=2, non_existent_property=10)
    assert "Some properties could not be set due to errors" in str(ex.value)
    # __notes__ is not standard in pytest exceptions, so we skip that assertion
    # but those that exist will still be written
    assert registry.thing_inst.integration_time == 120
    assert registry.thing_inst.trigger_mode == 2


def test_16_db_properties():
    """Test db operations for properties"""
    # req. 1. db operations are supported only at instance level
    with pytest.raises(AttributeError) as ex:
        Thing.properties.load_from_DB()
    assert "database operations are only supported at instance level" in str(ex.value)
    with pytest.raises(AttributeError) as ex:
        Thing.properties.get_from_DB()
    assert "database operations are only supported at instance level" in str(ex.value)


def test_17_inheritance_of_registries():
    """Test that registries are inherited properly"""
    # req. 1. subclass have more descriptors than parent class because our example Thing OceanOpticsSpectrometer
    # has defined its own actions, properties and events
    assert len(OceanOpticsSpectrometer.properties.descriptors) > len(Thing.properties.descriptors)
    assert len(OceanOpticsSpectrometer.actions.descriptors) > len(Thing.actions.descriptors)
    assert len(OceanOpticsSpectrometer.events.descriptors) > len(Thing.events.descriptors)


# """
# # Summary of tests and requirements:

# TestThing class:
# 1. Test id requirements:
#     - Instance name must be a string and cannot be changed after set.
#     - Valid and invalid IDs based on regex (r'[A-Za-z]+[A-Za-z_0-9\\-\\/]*').
# 2. Test logger setup:
#     - Logger must have remote access handler if remote_accessible_logger is True.
#     - Logger is created automatically if not provided.
# 3. Test state and state_machine setup:
#     - State property must be None when no state machine is present.
# 4. Test composition of subthings:
#     - Subthings must be a dictionary.
#     - Subthings are recomputed when accessed.
#     - Subthings must be instances of Thing and have the parent as owner.
#     - Name of subthing must match name of the attribute.
# 5. Test servers init:
#     - rpc_server and event_publisher must be None when not run().
#     - rpc_server and event_publisher must be instances of their respective classes when run().
# 6. Test thing model generation:
#     - Basic test to ensure nothing is fundamentally wrong.

# TestOceanOpticsSpectrometer class:
# 1. Test state and state_machine setup:
#     - State and state machine must be present because subclass has a state machine.
#     - State and state machine must be different for different instances.

# TestMetaclass class:
# 1. Test metaclass of Thing class:
#     - Metaclass must be ThingMeta for any Thing class.
# 2. Test registry creation and access:
#     - Registry attributes must be instances of their respective classes.
#     - New registries are not created on the fly and are same between accesses.
#     - Different subclasses have different registries.
#     - Registry attributes must be instances of their respective classes also for instances.
#     - Registries are not created on the fly and are same between accesses also for instances.
#     - Registries are not shared between instances.
#     - Registries are not shared between instances and their classes.

# TestRegistry class:
# 1. Test owner attribute:
#     - Owner attribute must be the class itself when accessed as class attribute.
#     - Owner attribute must be the instance for instance registries.
#     - Descriptor_object must be defined correctly and is a class.
# 2. Test descriptors access:
#     - Descriptors are instances of the descriptor object.
#     - Class level or instance level descriptors are same.
#     - Descriptors can be cleared.
# 3. Test dunders:
#     - __getitem__ must return the descriptor object.
#     - __contains__ must return True if the descriptor is present.
#     - __iter__ must return an iterator over the descriptors dictionary.
#     - __len__ must return the number of descriptors.
#     - Registries have their unique hashes.
# 4. Test bound objects:
#     - Number of bound objects must be equal to number of descriptors.
#     - Bound objects must be instances of bound instances.

# TestActionRegistry class:
# - Inherits tests from TestRegistry.

# TestEventRegistry class:
# - Inherits tests from TestRegistry.
# - Observables and change events are also descriptors.

# TestPropertiesRegistry class:
# - Inherits tests from TestRegistry.
# - Parameters that are subclass of Property are usually remote objects.
# - DB operations are supported only at instance level.
# """


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import ssl

from typing import Optional, Type  # noqa: F401

import aiomqtt
import structlog

from ...core import Thing as CoreThing
from ...param.parameters import ClassSelector, String
from ...td.interaction_affordance import EventAffordance, PropertyAffordance
from ...utils import get_current_async_loop
from ..repository import thing_repository
from ..server import BaseProtocolServer
from .config import RuntimeConfig
from .controllers import ThingDescriptionPublisher, TopicPublisher
from .services import ThingDescriptionService


class MQTTPublisher(BaseProtocolServer):
    """
    MQTT Publisher. All events and observable properties defined on the Thing will be published to MQTT topics
    with topic name "{thing id}/{event name}".

    For setting up an MQTT broker if one does not exist,
    see [infrastructure project](https://github.com/hololinked-dev/daq-system-infrastructure).
    """

    hostname = String(default="localhost")
    """The MQTT broker hostname"""

    ssl_context = ClassSelector(class_=ssl.SSLContext, allow_None=True, default=None)
    """The SSL context to use for secure connections, or None for no SSL"""

    config = ClassSelector(class_=RuntimeConfig, default=None, allow_None=True)  # type: RuntimeConfig
    """Runtime configuration for the MQTT publisher"""

    def __init__(
        self,
        hostname: str,
        port: int,
        username: str,
        password: str,
        qos: int = 1,
        things: Optional[list[CoreThing]] = None,
        config: Optional[dict] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        hostname: str
            The MQTT broker hostname
        port: int
            The MQTT broker port
        username: str
            The MQTT broker username
        password: str
            The MQTT broker password
        qos: int
            The (global) MQTT QoS level to use for publishing messages
        things: list[Thing]
            The `Thing`s that need to publish their events/properties to MQTT broker
        config: dict, optional
            Additional runtime configuration for the MQTT publisher, see `RuntimeConfig` object under
            `hololinked.server.mqtt.config`
        kwargs: dict
            Additional keyword arguments
        """
        endpoint = f"{self.hostname}{f':{self.port}' if self.port else ''}"

        default_config = dict(
            topic_publisher=kwargs.get("topic_publisher", TopicPublisher),
            thing_description_publisher=kwargs.get("thing_description_publisher", ThingDescriptionPublisher),
            thing_description_service=kwargs.get("thing_description_service", ThingDescriptionService),
            thing_repository=kwargs.get("thing_repository", thing_repository),
            qos=qos,
        )
        default_config.update(config or dict())
        config = RuntimeConfig(**default_config)

        self.hostname = hostname
        self.port = port
        self.config = config
        self.username = username
        self.password = password
        self.publishers = dict()  # type: dict[str, TopicPublisher]
        self.logger = kwargs.get("logger", structlog.get_logger()).bind(component="mqtt-publisher", hostname=endpoint)
        self.ssl_context = kwargs.get("ssl_context", None)
        self.add_things(*(things or []))

    async def start(self):
        """
        Sets up the MQTT client and starts publishing events from the `Thing`s.
        All events are dispatched to their own async tasks. This method returns and
        creates side-effects only & does not block. Use the `run()` method instead for a blocking call.
        """
        self.client = aiomqtt.Client(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            tls_context=self.ssl_context,
        )
        try:
            await self.client.__aenter__()
            endpoint = f"{self.hostname}{f':{self.port}' if self.port else ''}"
            self.logger.info(f"Connected to MQTT broker at {endpoint}")
        except aiomqtt.MqttReentrantError:
            pass
        # better to do later
        await self.setup()

    async def start_publishers(self, thing: CoreThing) -> None:
        """Start the publishers for a given `Thing`"""
        eventloop = get_current_async_loop()
        if not thing.rpc_server:
            raise ValueError(f"Thing {thing.id} is not associated with any RPC server")

        await self._instantiate_broker(server_id=thing.rpc_server.id, thing_id=thing.id, access_point="INPROC")
        TD = thing_repository[thing.id].TD

        for event_name in TD.get("events", {}).keys():
            event_affordance = EventAffordance.from_TD(event_name, TD)
            topic_publisher = self.config.topic_publisher(
                client=self.client,
                resource=event_affordance,
                logger=self.logger,
                config=self.config,
            )
            self.publishers[topic_publisher.topic] = topic_publisher
            eventloop.create_task(topic_publisher.publish())
            self.logger.info(f"MQTT will publish events for {event_name} of thing {thing.id}")
        for prop_name in TD.get("properties", {}).keys():
            property_affordance = PropertyAffordance.from_TD(prop_name, TD)
            if not property_affordance.observable:
                continue
            topic_publisher = self.config.topic_publisher(
                client=self.client,
                resource=property_affordance,
                logger=self.logger,
                config=self.config,
            )
            self.publishers[topic_publisher.topic] = topic_publisher
            eventloop.create_task(topic_publisher.publish())
            self.logger.info(f"MQTT will publish observable property changes for {prop_name} of thing {thing.id}")
        # TD publisher
        td_publisher = self.config.thing_description_publisher(
            client=self.client,
            logger=self.logger,
            ZMQ_TD=TD,
            config=self.config,
        )
        self.publishers[td_publisher.topic] = td_publisher
        eventloop.create_task(td_publisher.publish(TD))

    async def setup(self) -> None:
        """Setup MQTT publishers per `Thing` post connection to broker"""
        eventloop = get_current_async_loop()
        for thing in self.things:
            eventloop.create_task(self.start_publishers(thing))

    def stop(self):
        """stop publishing, the client is not closed automatically"""
        for publisher in self.publishers.values():
            publisher.stop()



import inspect

from typing import Any, ClassVar

from pydantic import BaseModel


class Schema(BaseModel):
    """
    Base pydantic model for all WoT schema components (as in, parts within the schema).
    Call `model_dump` or `json` method to get the JSON representation of the schema.
    """

    skip_keys: ClassVar = []  # override this to skip some dataclass attributes in the schema

    def model_dump(self, **kwargs) -> dict[str, Any]:
        """Return the JSON representation of the schema"""
        # we need to override this to work with our JSON serializer
        kwargs["mode"] = "json"
        kwargs["by_alias"] = True
        kwargs["exclude_unset"] = True
        kwargs["exclude"] = [
            "instance",
            "skip_keys",
            "skip_properties",
            "skip_actions",
            "skip_events",
            "ignore_errors",
            "allow_loose_schema",
        ]
        return super().model_dump(**kwargs)

    def json(self) -> dict[str, Any]:
        """same as model_dump"""
        return self.model_dump()

    @classmethod
    def format_doc(cls, doc: str):
        """strip tabs, newlines, whitespaces etc. to format the docstring nicely"""
        doc = inspect.cleandoc(doc)
        # Remove everything after "Parameters\n-----" if present (when using numpydoc)
        marker = "Parameters\n-----"
        idx = doc.find(marker)
        if idx != -1:
            doc = doc[:idx]
        doc = doc.replace("\n", "").replace("\t", " ").lstrip().rstrip()
        return doc



import pytest

from things import TestThing

from hololinked.serializers import Serializers
from hololinked.serializers.serializers import BaseSerializer


class YAMLSerializer(BaseSerializer):
    """just a dummy, does not really serialize to YAML"""

    @property
    def content_type(self):
        return "application/yaml"


@pytest.fixture(scope="module")
def yaml_serializer() -> BaseSerializer:
    # test register a new serializer with content type
    return YAMLSerializer()


def test_01_singleton():
    """Test the singleton nature of the Serializers class."""

    serializers = Serializers()
    assert serializers == Serializers()
    assert Serializers != Serializers()
    assert isinstance(serializers, Serializers)
    # all are class attributes
    assert serializers.json == Serializers.json
    assert serializers.pickle == Serializers.pickle
    assert serializers.msgpack == Serializers.msgpack
    assert serializers.content_types == Serializers.content_types
    assert serializers.object_content_type_map == Serializers.object_content_type_map
    assert serializers.object_serializer_map == Serializers.object_serializer_map
    assert serializers.protocol_serializer_map == Serializers.protocol_serializer_map
    # check existing serializers are all instances of BaseSerializer
    for name, serializer in Serializers.content_types.items():
        assert isinstance(serializer, BaseSerializer)
    # check default serializer, given that we know its JSON at least for the current test
    assert serializers.default == Serializers.json
    assert serializers.default == Serializers.default
    assert serializers.default == Serializers().json
    assert serializers.default == Serializers().default
    # check default content type, given that we know its JSON at least for the current test
    assert serializers.default_content_type == Serializers.json.content_type
    # change default to pickle and check if it is set correctly
    # serializers.default = serializers.pickle
    # self.assertEqual(serializers.default, Serializers.pickle)
    # self.assertEqual(Serializers().default, Serializers.pickle)


def test_02_protocol_registration(yaml_serializer: BaseSerializer):
    """i.e. test if a new serializer (protocol) can be registered"""

    # get existing number of serializers
    num_serializers = len(Serializers.content_types)

    # test register a new serializer
    base_serializer = BaseSerializer()
    # register with name
    with pytest.warns(UserWarning):
        Serializers.register(base_serializer, "base")
    # user warning because content type property is not defined
    # above is same as Serializers.register(base_serializer, 'base')

    # check if name became a class attribute and name can be accessed as an attribute
    assert "base" in Serializers
    assert Serializers.base == base_serializer
    assert Serializers().base == base_serializer
    # we dont support getitem at instance level yet so we cannot test assertIn

    # since a content type is not set, it should not be in the content types
    assert base_serializer not in Serializers.content_types.values()
    # so the length of content types should be the same
    assert len(Serializers.content_types) == num_serializers

    # register with name
    Serializers.register(yaml_serializer, "yaml")
    # check if name became a class attribute and name can be accessed as an attribute
    assert "yaml" in Serializers
    assert Serializers.yaml == yaml_serializer
    assert Serializers().yaml == yaml_serializer
    # we dont support getitem at instance level yet

    # since a content type is set, it should be in the content types
    assert yaml_serializer.content_type in Serializers.content_types.keys()
    assert yaml_serializer in Serializers.content_types.values()
    # so the length of content types should have increased by 1
    assert len(Serializers.content_types) == num_serializers + 1


def test_03_registration_for_objects():
    """i.e. test if a new serializer can be registered for a specific property, action or event"""
    Serializers.register_content_type_for_object(TestThing.base_property, "application/x-pickle")
    Serializers.register_content_type_for_object(TestThing.action_echo, "application/msgpack")
    Serializers.register_content_type_for_object(TestThing.test_event, "application/yaml")

    assert Serializers.for_object(None, "TestThing", "action_echo") == Serializers.msgpack
    assert Serializers.for_object(None, "TestThing", "base_property") == Serializers.pickle
    assert Serializers.for_object(None, "TestThing", "test_event") == Serializers.yaml
    assert Serializers.for_object(None, "TestThing", "test_unknown_property") == Serializers.default


def test_04_registration_for_objects_by_name():
    Serializers.register_content_type_for_object_per_thing_instance("test_thing", "base_property", "application/yaml")
    assert isinstance(Serializers.for_object("test_thing", None, "base_property"), YAMLSerializer)


def test_05_registration_dict():
    """test the dictionary where all serializers are stored"""
    # depends on test 3
    assert "test_thing" in Serializers.object_content_type_map
    assert "base_property" in Serializers.object_content_type_map["test_thing"]
    assert Serializers.object_content_type_map["test_thing"]["base_property"] == "application/yaml"
    assert Serializers.object_content_type_map["test_thing"]["base_property"] == "application/yaml"

    assert "action_echo" in Serializers.object_content_type_map["TestThing"]
    assert Serializers.object_content_type_map["TestThing"]["action_echo"] == "application/msgpack"
    assert "test_event" in Serializers.object_content_type_map["TestThing"]
    assert Serializers.object_content_type_map["TestThing"]["test_event"] == "application/yaml"


def test_06_retrieval():
    # added in previous tests
    assert isinstance(Serializers.for_object("test_thing", None, "base_property"), YAMLSerializer)
    # unknown object should retrieve the default serializer
    assert Serializers.for_object("test_thing", None, "test_unknown_property") == Serializers.default
    # unknown thing should retrieve the default serializer
    assert Serializers.for_object("test_unknown_thing", None, "base_property") == Serializers.default


def test_07_set_default():
    """test setting the default serializer"""
    # get existing default
    old_default = Serializers.default
    # set new default and check if default is set
    Serializers.default = Serializers.yaml
    assert Serializers.default == Serializers.yaml
    test_06_retrieval()  # check if retrieval is consistent with default
    # reset default and check if default is reset
    Serializers.default = old_default
    assert Serializers.default == old_default
    assert Serializers.default == Serializers.json  # because we know its JSON


def test_08_unknown_content_types():
    """test registration of unknown content types for objects"""
    Serializers.register_content_type_for_object(TestThing.number_prop, "application/unknown")
    assert Serializers.for_object(None, "TestThing", "number_prop") is None
    assert Serializers.get_content_type_for_object(None, "TestThing", "number_prop") == "application/unknown"


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import copy

from enum import Enum
from typing import Any, ClassVar, Optional

from pydantic import BaseModel, ConfigDict, RootModel

from ..constants import JSON, ResourceTypes
from ..core.actions import Action
from ..core.events import Event
from ..core.property import Property
from ..core.thing import Thing, ThingMeta
from ..utils import issubklass
from .base import Schema
from .data_schema import DataSchema
from .forms import Form
from .pydantic_extensions import type_to_dataschema
from .utils import get_summary


class InteractionAffordance(Schema):
    """
    Implements schema information common to all interaction affordances.

    [Specification Definitions](https://www.w3.org/TR/wot-thing-description11/#interactionaffordance) <br>
    [UML Diagram](https://docs.hololinked.dev/UML/PDF/InteractionAffordance.pdf) <br>
    """

    title: Optional[str] = None
    titles: Optional[dict[str, str]] = None
    description: Optional[str] = None
    descriptions: Optional[dict[str, str]] = None
    forms: Optional[list[Form]] = None
    # uri variables

    _custom_schema_generators: ClassVar = dict()
    model_config = ConfigDict(extra="allow")

    def __init__(self):
        super().__init__()
        self._name = None
        self._objekt = None
        self._thing_id = None
        self._thing_cls = None
        self._owner = None

    @property
    def what(self) -> Enum:
        """Whether it is a property, action or event"""
        raise NotImplementedError("Unknown interaction affordance - implement in subclass of InteractionAffordance")

    @property
    def owner(self) -> Thing:
        """
        Owning `Thing` instance or `Thing` class of the interaction affordance.
        Depends on how this object was created, whether using an instance or a class.
        """
        return self._owner

    @owner.setter
    def owner(self, value):
        if self._owner is not None:
            raise ValueError(
                f"owner is already set for this {self.what.name.lower()} affordance, "
                + "recreate the affordance to change owner"
            )
        if not isinstance(value, (Thing, ThingMeta)):
            raise TypeError(f"owner must be instance of Thing, given type {type(value)}")
        self._owner = value
        if isinstance(value, Thing):
            self._thing_cls = value.__class__
            self._thing_id = value.id
        elif isinstance(value, ThingMeta):
            self._thing_cls = value

    @property
    def objekt(self) -> Property | Action | Event:
        """Object instance of the interaction affordance - `Property`, `Action` or `Event`"""
        return self._objekt

    @objekt.setter
    def objekt(self, value: Property | Action | Event) -> None:
        """Set the object instance of the interaction affordance - `Property`, `Action` or `Event`"""
        if self._objekt is not None:
            raise ValueError(
                f"object is already set for this {self.what.name.lower()} affordance, "
                + "recreate the affordance to change objekt"
            )
        if not (
            (self.__class__.__name__.startswith("Property") and isinstance(value, Property))
            or (self.__class__.__name__.startswith("Action") and isinstance(value, Action))
            or (self.__class__.__name__.startswith("Event") and isinstance(value, Event))
        ):
            if not isinstance(value, (Property, Action, Event)):
                raise TypeError(f"objekt must be instance of Property, Action or Event, given type {type(value)}")
            raise ValueError(
                f"provide only corresponding object for {self.__class__.__name__}, "
                + f"given object {value.__class__.__name__}"
            )
        self._objekt = value
        self._name = value.name

    @property
    def name(self) -> str:
        """Name of the interaction affordance used as key in the TD"""
        return self._name

    @property
    def thing_id(self) -> str | None:
        """ID of the `Thing` instance owning the interaction affordance, if available, otherwise None"""
        return self._thing_id

    @property
    def thing_cls(self) -> ThingMeta:
        """`Thing` class owning the interaction affordance"""
        return self._thing_cls

    def build(self) -> None:
        """populate the fields of the schema for the specific interaction affordance"""
        raise NotImplementedError("build must be implemented in subclass of InteractionAffordance")

    def retrieve_form(self, op: str, default: Any = None) -> Form:
        """
        retrieve form for a certain operation, return default if not found

        Parameters
        ----------
        op: str
            operation for which the form is to be retrieved
        default: Any, optional
            default value to return if form is not found, by default None.
            One can make use of a sensible default value for one's logic.

        Returns
        -------
        dict[str, Any]
            JSON representation of the form
        """
        if self.forms is None:
            return default
        for form in self.forms:
            if form.op == op:
                return form
        return default

    def pop_form(self, op: str, default: Any = None) -> Form:
        """
        retrieve and remove form for a certain operation, return default if not found

        Parameters
        ----------
        op: str
            operation for which the form is to be retrieved
        default: Any, optional
            default value to return if form is not found, by default None.
            One can make use of a sensible default value for one's logic.

        Returns
        -------
        dict[str, Any]
            JSON representation of the form
        """
        if self.forms is None:
            return default
        for i, form in enumerate(self.forms):
            if form.op == op:
                return self.forms.pop(i)
        return default

    @classmethod
    def generate(
        cls,
        interaction: Property | Action | Event,
        owner: Thing,
    ) -> "PropertyAffordance | ActionAffordance | EventAffordance":
        """
        build the schema for the specific interaction affordance as an instance of this class.
        Use the `json()` method to get the JSON representation of the schema.

        Note that this method is different from build() method as its supposed to be used as a classmethod
        to create an instance. Although, it internally calls build(), and some additional steps are included.

        Parameters
        ----------
        interaction: Property | Action | Event
            interaction object for which the schema is to be built
        owner: Thing
            owner of the interaction affordance

        Returns
        -------
        "PropertyAffordance | ActionAffordance | EventAffordance"
        """
        raise NotImplementedError("generate_schema must be implemented in subclass of InteractionAffordance")

    @classmethod
    def from_TD(cls, name: str, TD: JSON) -> "PropertyAffordance | ActionAffordance | EventAffordance":
        """
        populate the schema from the TD and return it as an instance of this class.

        Parameters
        ----------
        name: str
            name of the interaction affordance used as key in the TD
        TD: JSON
            Thing Description JSON dictionary (the entire one, not just the component of the affordance)

        Returns
        -------
        "PropertyAffordance | ActionAffordance | EventAffordance"
        """
        if cls == PropertyAffordance:
            affordance_name = "properties"
        elif cls == ActionAffordance:
            affordance_name = "actions"
        elif cls == EventAffordance:
            affordance_name = "events"
        else:
            raise ValueError(f"unknown affordance type - {cls}, cannot create object from TD")
        affordance_json = TD[affordance_name][name]  # type: dict[str, JSON]
        affordance = cls()
        for field in cls.model_fields:
            if field in affordance_json:
                if field == "forms":
                    affordance.forms = [Form.from_TD(form) for form in affordance_json[field]]
                else:
                    setattr(affordance, field, affordance_json[field])
        affordance._name = name
        affordance._thing_id = TD["id"]
        return affordance

    @classmethod
    def register_descriptor(
        cls,
        descriptor: Property | Action | Event,
        schema_generator: "InteractionAffordance",
    ) -> None:
        """register a custom schema generator for a descriptor"""
        if not isinstance(descriptor, (Property, Action, Event)):
            raise TypeError(
                "custom schema generator can also be registered for Property." + f" Given type {type(descriptor)}"
            )
        if not isinstance(schema_generator, InteractionAffordance):
            raise TypeError(
                "schema generator for Property must be subclass of PropertyAfforance. "
                + f"Given type {type(schema_generator)}"
            )
        InteractionAffordance._custom_schema_generators[descriptor] = schema_generator

    def build_non_compliant_metadata(self) -> None:
        """If by chance, there is additional non standard metadata to be added, they can be added here"""
        pass

    def override_defaults(self, **kwargs):
        """
        Override default values with provided keyword arguments, especially thing_id, owner name, object name etc.
        Any logic to trigger side effects while setting those values should be handled here.
        """
        for key, value in kwargs.items():
            if key == "name":
                self._name = value
            elif key == "thing_id":
                self._thing_id = value
            elif key == "owner":
                self._owner = value
            elif key == "thing_cls":
                self._thing_cls = value
            elif hasattr(self, key) or key in self.model_fields:
                setattr(self, key, value)

    def __hash__(self):
        return hash(
            self.thing_id if self.thing_id else "" + self.thing_cls.__name__ if self.thing_cls else "" + self.name
        )

    def __str__(self):
        if self.thing_cls:
            return f"{self.__class__.__name__}({self.thing_cls.__name__}({self.thing_id}).{self.name})"
        return f"{self.__class__.__name__}({self.name} of {self.thing_id})"

    def __eq__(self, value):
        if not isinstance(value, self.__class__):
            return False
        if self.thing_id is None or value.thing_id is None:
            if self.owner is None or value.owner is None:
                # cannot determine anymore
                return False
            # basically you need to have an owner for the interaction affordance
            # and a name to determine its equality. We should never check the owner
            # by the name, but by the object, otherwise the equality cannot be gauranteed
            if (self.owner == value.owner or self.thing_cls == value.thing_cls) and self.name == value.name:
                return True
            return False
        return self.thing_id == value.thing_id and self.name == value.name

    def __deepcopy__(self, memo):
        if self.__class__ == PropertyAffordance:
            result = PropertyAffordance()
        elif self.__class__ == ActionAffordance:
            result = ActionAffordance()
        elif self.__class__ == EventAffordance:
            result = EventAffordance()
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ("_owner", "_thing_cls", "_objekt"):
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove possibly unpicklable entries
        if "_owner" in state:
            del state["_owner"]
        if "_thing_cls" in state:
            del state["_thing_cls"]
        if "_objekt" in state:
            del state["_objekt"]
        return state


class PropertyAffordance(DataSchema, InteractionAffordance):
    """
    Implements property affordance schema from `Property` descriptor object.

    [Schema](https://www.w3.org/TR/wot-thing-description11/#propertyaffordance) <br>
    [UML Diagram](https://docs.hololinked.dev/UML/PDF/InteractionAffordance.pdf) <br>
    """

    # [Supported Fields]() <br>
    observable: Optional[bool] = None

    def __init__(self):
        super().__init__()

    @property
    def what(self) -> Enum:
        return ResourceTypes.PROPERTY

    def build(self) -> None:
        property = self.objekt
        self.ds_build_from_property(property)
        if property.observable:
            self.observable = property.observable

    @classmethod
    def generate(cls, property, owner=None):
        if not isinstance(property, Property):
            raise TypeError(f"property must be instance of Property, given type {type(property)}")
        affordance = PropertyAffordance()
        affordance.owner = owner
        affordance.objekt = property
        affordance.build()
        affordance.build_non_compliant_metadata()
        return affordance


class ActionAffordance(InteractionAffordance):
    """
    creates action affordance schema from actions (or methods).

    [Schema](https://www.w3.org/TR/wot-thing-description11/#actionaffordance) <br>
    [UML Diagram](https://docs.hololinked.dev/UML/PDF/InteractionAffordance.pdf) <br>
    """

    # [Supported Fields]() <br>
    input: JSON = None
    output: JSON = None
    safe: bool = None
    idempotent: bool = None
    synchronous: bool = None

    def __init__(self):
        super().__init__()

    @property
    def what(self):
        return ResourceTypes.ACTION

    def build(self) -> None:
        action = self.objekt  # type: Action
        if action.obj.__doc__:
            title = get_summary(action.obj.__doc__)
            description = self.format_doc(action.obj.__doc__)
            if title and not description.startswith(title):
                self.title = title
                self.description = description
            else:
                self.description = description
        if action.execution_info.argument_schema:
            if isinstance(action.execution_info.argument_schema, dict):
                self.input = action.execution_info.argument_schema
            elif issubklass(action.execution_info.argument_schema, (BaseModel, RootModel)):
                self.input = type_to_dataschema(action.execution_info.argument_schema)
            else:
                raise ValueError(
                    f"unknown schema definition for action input, given type: {type(action.execution_info.argument_schema)}"
                )
        if action.execution_info.return_value_schema:
            if isinstance(action.execution_info.return_value_schema, dict):
                self.output = action.execution_info.return_value_schema
            elif issubklass(action.execution_info.return_value_schema, (BaseModel, RootModel)):
                self.output = type_to_dataschema(action.execution_info.return_value_schema)
            else:
                raise ValueError(
                    f"unknown schema definition for action output, given type: {type(action.execution_info.return_value_schema)}"
                )
        if (
            not (
                hasattr(self.owner, "state_machine")
                and self.owner.state_machine is not None
                and self.owner.state_machine.contains_object(action)
            )
            and action.execution_info.idempotent
        ):
            self.idempotent = action.execution_info.idempotent
        if action.execution_info.synchronous:
            self.synchronous = action.execution_info.synchronous
        if action.execution_info.safe:
            self.safe = action.execution_info.safe

    @classmethod
    def generate(cls, action: Action, owner, **kwargs) -> "ActionAffordance":
        if not isinstance(action, Action):
            raise TypeError(f"action must be instance of Action, given type {type(action)}")
        affordance = ActionAffordance()
        affordance.owner = owner
        affordance.objekt = action
        affordance.build()
        affordance.build_non_compliant_metadata()
        return affordance


class EventAffordance(InteractionAffordance):
    """
    creates event affordance schema from events.

    [Schema](https://www.w3.org/TR/wot-thing-description11/#eventaffordance) <br>
    [UML Diagram](https://docs.hololinked.dev/UML/PDF/InteractionAffordance.pdf) <br>
    """

    # [Supported Fields]() <br>
    subscription: str = None
    data: JSON = None

    def __init__(self):
        super().__init__()

    @property
    def what(self):
        return ResourceTypes.EVENT

    def build(self) -> None:
        event = self.objekt  # type: Event
        if event.__doc__:
            title = get_summary(event.doc)
            description = self.format_doc(event.doc)
            if title and not description.startswith(title):
                self.title = title
                self.description = description
            else:
                self.description = description
        if event.schema:
            if isinstance(event.schema, dict):
                self.data = event.schema
            elif issubklass(event.schema, (BaseModel, RootModel)):
                self.data = type_to_dataschema(event.schema)
            else:
                raise ValueError(f"unknown schema definition for event data, given type: {type(event.schema)}")

    @classmethod
    def generate(cls, event: Event, owner, **kwargs) -> "EventAffordance":
        if not isinstance(event, Event):
            raise TypeError(f"event must be instance of Event, given type {type(event)}")
        affordance = EventAffordance()
        affordance.owner = owner
        affordance.objekt = event
        affordance.build()
        affordance.build_non_compliant_metadata()
        return affordance



from typing import Annotated, Any

from pydantic import BaseModel, Field

from ..repository import thing_repository
from .controllers import ThingDescriptionPublisher, TopicPublisher
from .services import ThingDescriptionService


class RuntimeConfig(BaseModel):
    """
    Runtime configuration for MQTT publishers, initialized in `MQTTPublisher` object.
    Pass the attributes of this class as a dictionary to the `config` argument of `MQTTPublisher`.
    """

    qos: Annotated[int, Field(ge=0, le=2)] = 1
    """The (global) MQTT QoS level to use for publishing messages"""

    topic_publisher: type[TopicPublisher] | Any = TopicPublisher
    """handler class to be used for publishing to topics (global)"""
    thing_description_publisher: type[ThingDescriptionPublisher] | Any = ThingDescriptionPublisher
    """handler class to be used for publishing thing descriptions"""

    thing_description_service: type[ThingDescriptionService] | Any = ThingDescriptionService
    """Thing Description generation service, used by `ThingDescriptionPublisher` to generate the Thing Description"""

    thing_repository: Any = thing_repository  # type: dict[str, BrokerThing]
    """repository layer `Thing`s to be used by the MQTT publishers"""



import base64
import threading
import time

import httpx

from pydantic import BaseModel, PrivateAttr


class BasicSecurity(BaseModel):
    """
    Basic Security Scheme with username and password.
    The credentials are added into the `Authorization` header.
    """

    http_header_name: str = "Authorization"

    _credentials: str = PrivateAttr()

    def __init__(self, username: str, password: str, use_base64: bool = True):
        """
        Parameters
        ----------
        username: str
            The username for basic authentication
        password: str
            The password for basic authentication
        use_base64: bool
            Whether to encode the credentials in base64, by default True
        """
        super().__init__()
        credentials = f"{username}:{password}"
        if use_base64:
            credentials = base64.b64encode(credentials.encode("utf-8")).decode("utf-8")
        self._credentials = f"Basic {credentials}"

    @property
    def http_header(self) -> str:
        """Value for the Authorization header"""
        return self._credentials


class APIKeySecurity(BaseModel):
    """
    API Key Security Scheme.
    The API key is added into a header named `X-API-Key`.
    """

    value: str
    http_header_name: str = "X-API-Key"

    @property
    def http_header(self) -> str:
        return self.value


class ROPC(BaseModel):
    access_token: str
    scope: str
    refresh_token: str | None = None
    expires_in: int | None = None
    token_type: str | None = None
    id_token: str | None = None


class OAuthDirectAccessGrant(BaseModel):
    """
    OAuth2 Direct Access Grant Security Scheme.
    Implements Resource Owner Password Credentials (ROPC) flow.
    """

    token_endpoint: str
    client_id: str
    client_secret: str | None = None
    revocation_endpoint: str | None = None

    username: str
    password: str
    scope: str | list[str] = "openid"
    grant_type: str = "password"

    def __init__(
        self,
        username: str,
        password: str,
        oidc_config_url: str = None,
        scope: str | list[str] = "openid",
        verify_ssl: bool = True,
        **kwargs,
    ):
        token_endpoint = kwargs.get("token_endpoint", None)
        client_id = kwargs.get("client_id", None)
        client_secret = kwargs.get("client_secret", None)
        revocation_endpoint = kwargs.get("revocation_endpoint", None)
        if oidc_config_url:
            with httpx.Client(timeout=10.0, verify=verify_ssl) as client:
                response = client.get(oidc_config_url)
                response.raise_for_status()
                oidc_config = response.json()
                token_endpoint = oidc_config["token_endpoint"]
                revocation_endpoint = oidc_config.get("revocation_endpoint", None)
        elif not token_endpoint:
            raise ValueError("Either 'oidc_config_url' or 'token_endpoint' must be provided")
        super().__init__(
            token_endpoint=token_endpoint,
            client_id=client_id,
            client_secret=client_secret,
            revocation_endpoint=revocation_endpoint,
            username=username,
            password=password,
            scope=scope,
        )


class OAuth2Security:
    """
    OAuth2 Security Scheme, Currently only supports Resource Owner Password Credentials (ROPC) flow.
    Please implement other flows on your own for applications with a web interface.
    """

    http_header_name: str = "Authorization"

    def __init__(
        self,
        oidc_settings: OAuthDirectAccessGrant,
        req_rep_sync_client: httpx.Client,
        req_rep_async_client: httpx.AsyncClient,
        refresh_interval_fraction: float = 0.75,
    ) -> None:
        self._oidc_settings = oidc_settings
        self._req_rep_async_client = req_rep_async_client
        self._req_rep_sync_client = req_rep_sync_client
        self.tokens = None
        self._refresh_thread = None
        self._refresh_lock = threading.Lock()
        self._refresh = True
        self._refresh_interval_fraction = refresh_interval_fraction

    @property
    def http_header(self) -> str:
        if not self.tokens:
            return ""
        try:
            self._refresh_lock.acquire()
            return f"Bearer {self.tokens.access_token}"
        finally:
            self._refresh_lock.release()

    def login(self) -> None:
        """login with username and password and obtain tokens"""
        try:
            self._refresh_lock.acquire()
            body = dict(
                grant_type=self._oidc_settings.grant_type,
                client_id=self._oidc_settings.client_id,
                scope=self._oidc_settings.scope,
                username=self._oidc_settings.username,
                password=self._oidc_settings.password,
            )
            if self._oidc_settings.client_secret:
                body["client_secret"] = self._oidc_settings.client_secret
            response = self._req_rep_sync_client.post(
                self._oidc_settings.token_endpoint,
                data=body,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            response.raise_for_status()
            self.tokens = ROPC(
                access_token=response.json().get("access_token"),
                refresh_token=response.json().get("refresh_token"),
                expires_in=response.json().get("expires_in"),
                id_token=response.json().get("id_token"),
                scope=response.json().get("scope"),
                token_type=response.json().get("token_type"),
            )
            if self._refresh_thread and self._refresh_thread.is_alive():
                return
            if not self.tokens.refresh_token or not self.tokens.expires_in:
                return
            self._refresh_thread = threading.Thread(target=self._refresh_tokens_in_background, daemon=True)
            self._refresh_thread.start()
        finally:
            self._refresh_lock.release()

    def logout(self) -> None:
        """logout and invalidate tokens"""
        body = dict(
            client_id=self._oidc_settings.client_id,
            token=self.tokens.refresh_token if self.tokens.refresh_token else self.tokens.access_token,
            token_type_hint="refresh_token" if self.tokens.refresh_token else "access_token",
        )
        if self._oidc_settings.client_secret:
            body["client_secret"] = self._oidc_settings.client_secret
        response = self._req_rep_sync_client.post(
            self._oidc_settings.revocation_endpoint,
            data=body,
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        response.raise_for_status()
        self.tokens = None
        self._refresh = False

    def refresh_tokens(self) -> None:
        """refresh tokens, even forcibly by relogin if necessary"""
        try:
            self._refresh_lock.acquire()
            body = dict(
                grant_type="refresh_token",
                client_id=self._oidc_settings.client_id,
                refresh_token=self.tokens.refresh_token,
            )
            if self._oidc_settings.client_secret:
                body["client_secret"] = self._oidc_settings.client_secret
            response = self._req_rep_sync_client.post(
                self._oidc_settings.token_endpoint,
                data=body,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            response.raise_for_status()
            self.tokens = ROPC(**response.json())
        except httpx.HTTPStatusError:
            self._refresh_lock.release()
            self.login()
        finally:
            self._refresh_lock.release()

    def _refresh_tokens_in_background(self) -> None:
        """background thread to refresh tokens periodically"""
        time.sleep(int(0.75 * self.tokens.expires_in))
        while self._refresh:
            self.refresh_tokens()
            time.sleep(int(self._refresh_interval_fraction * self.tokens.expires_in))
